"""
This module provides a function to easily and efficiently plot the rows of a ragged array.
"""

import numpy as np
import pandas as pd
import xarray as xr

from clouddrift.ragged import rowsize_to_index, segment


def plot_ragged(
    ax,
    longitude: list | np.ndarray | pd.Series | xr.DataArray,
    latitude: list | np.ndarray | pd.Series | xr.DataArray,
    rowsize: list | np.ndarray | pd.Series | xr.DataArray,
    *args,
    colors: list | np.ndarray | pd.Series | xr.DataArray | None = None,
    tolerance: float | int = 180,
    **kwargs,
):
    """Plot individually the rows of a ragged array dataset on a Matplotlib Axes
    or a Cartopy GeoAxes object ``ax``.

    This function wraps Matplotlib's ``plot`` function (``plt.plot``) and
    ``LineCollection`` (``matplotlib.collections``) to efficiently plot
    the rows of a ragged array dataset.

    Parameters
    ----------
    ax: matplotlib.axes.Axes or cartopy.mpl.geoaxes.GeoAxes
        Axis to plot on.
    longitude : array-like
        Longitude sequence. Unidimensional array input.
    latitude : array-like
        Latitude sequence. Unidimensional array input.
    rowsize : list
        List of integers specifying the number of data points in each row.
    *args : tuple
        Additional arguments to pass to ``ax.plot``.
    colors : array-like
        Values to map on the current colormap. If ``colors`` is the shape of ``rowsize``,
        the data points of each row are uniformly colored according to the color value for the row.
        If ``colors`` is the shape of ``longitude`` and ``latitude``, the data points are colored according
        to the color value for each data point.
    tolerance : float
        Longitude tolerance gap between data points (in degrees) for segmenting rows.
        For periodic domains, the tolerance parameter should be set to the maximum allowed gap
        between data points. Defaults to 180.
    **kwargs : dict
        Additional keyword arguments to pass to ``ax.plot``.

    Returns
    -------
    list of matplotlib.lines.Line2D or matplotlib.collections.LineCollection
        The plotted lines or line collection. Can be used to set a colorbar
        after plotting or extract information from the lines.

    Examples
    --------

    Load 100 trajectories from the gdp1h dataset for the examples.

    >>> from clouddrift import datasets
    >>> from clouddrift.ragged import subset
    >>> from clouddrift.plotting import plot_ragged
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from mpl_toolkits.axes_grid1 import make_axes_locatable
    >>> ds = datasets.gdp1h()
    >>> ds = subset(ds, {"id": ds.id[:100].values}).load()

    Plot the trajectories, assigning a different color to each trajectory:

    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(1, 1, 1)
    >>> l = plot_ragged(
    >>>     ax,
    >>>     ds.lon,
    >>>     ds.lat,
    >>>     ds.rowsize,
    >>>     colors=np.arange(len(ds.rowsize))
    >>> )
    >>> divider = make_axes_locatable(ax)
    >>> cax = divider.append_axes('right', size='3%', pad=0.05)
    >>> fig.colorbar(l, cax=cax)

    To plot the same trajectories, but assigning a different color to each
    data point based on a transformation of the time variable mapped onto
    the ``inferno`` colormap:

    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(1, 1, 1)
    >>> time = [v.astype(np.int64) / 86400 / 1e9 for v in ds.time.values]
    >>> l = plot_ragged(
    >>>     ax,
    >>>     ds.lon,
    >>>     ds.lat,
    >>>     ds.rowsize,
    >>>     colors=np.floor(time),
    >>>     cmap="inferno"
    >>> )
    >>> divider = make_axes_locatable(ax)
    >>> cax = divider.append_axes('right', size="3%", pad=0.05)
    >>> fig.colorbar(l, cax=cax)

    Finally, to plot the same trajectories, but using a cartopy
    projection:

    >>> import cartopy.crs as ccrs
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(1, 1, 1, projection=ccrs.Mollweide())
    >>> l = plot_ragged(
    >>>     ax,
    >>>     ds.lon,
    >>>     ds.lat,
    >>>     ds.rowsize,
    >>>     colors=np.arange(len(ds.rowsize)),
    >>>     transform=ccrs.PlateCarree(),
    >>>     cmap="Blues",
    >>> )
    >>> ax.set_extent([-180, 180, -90, 90])
    >>> ax.coastlines()
    >>> ax.gridlines(draw_labels=True)
    >>> divider = make_axes_locatable(ax)
    >>> cax = divider.append_axes('right', size="3%", pad=0.25, axes_class=plt.Axes)
    >>> fig.colorbar(l, cax=cax)

    Raises
    ------
    ValueError
        If longitude and latitude arrays do not have the same shape.
        If colors do not have the same shape as longitude and latitude arrays or rowsize.
        If ax is not a matplotlib Axes or GeoAxes object.
        If ax is a GeoAxes object and the transform keyword argument is not provided.

    ImportError
        If matplotlib is not installed.
        If the axis is a GeoAxes object and cartopy is not installed.
    """

    # optional dependency
    try:
        import matplotlib.colors as mcolors
        import matplotlib.pyplot as plt
        from matplotlib import cm
        from matplotlib.collections import LineCollection
    except ImportError:
        raise ImportError("missing optional dependency 'matplotlib'")

    if hasattr(ax, "coastlines"):  # check if GeoAxes without cartopy
        try:
            from cartopy.mpl.geoaxes import GeoAxes

            if isinstance(ax, GeoAxes) and not kwargs.get("transform"):
                raise ValueError(
                    "For GeoAxes, the transform keyword argument must be provided."
                )
        except ImportError:
            raise ImportError("missing optional dependency 'cartopy'")
    elif not isinstance(ax, plt.Axes):
        raise ValueError("ax must be either: plt.Axes or GeoAxes.")

    if np.sum(rowsize) != len(longitude):
        raise ValueError("The sum of rowsize must equal the length of lon and lat.")

    if len(longitude) != len(latitude):
        raise ValueError("lon and lat must have the same length.")

    if colors is None:
        colors = np.arange(len(rowsize))
    elif colors is not None and (len(colors) not in [len(longitude), len(rowsize)]):
        raise ValueError("shape colors must match the shape of lon/lat or rowsize.")

    # define a colormap
    if isinstance(cmap := kwargs.pop("cmap", cm.viridis), str):
        cmap = plt.get_cmap(cmap)

    # define a normalization obtain uniform colors
    # for the sequence of lines or LineCollection
    norm = kwargs.pop(
        "norm", mcolors.Normalize(vmin=np.nanmin(colors), vmax=np.nanmax(colors))
    )

    # create Mappable for colorbar
    cb = plt.cm.ScalarMappable(norm=norm, cmap=cmap)

    mpl_plot = True if colors is None or len(colors) == len(rowsize) else False
    traj_idx = rowsize_to_index(rowsize)

    for i in range(len(rowsize)):
        lon_i, lat_i = (
            longitude[traj_idx[i] : traj_idx[i + 1]],
            latitude[traj_idx[i] : traj_idx[i + 1]],
        )

        start = 0
        for length in segment(lon_i, tolerance, rowsize=segment(lon_i, -tolerance)):
            end = start + length

            if mpl_plot:
                ax.plot(
                    lon_i[start:end],
                    lat_i[start:end],
                    c=cmap(norm(colors[i])) if colors is not None else None,
                    *args,
                    **kwargs,
                )
            else:
                colors_i = colors[traj_idx[i] : traj_idx[i + 1]]
                segments = np.column_stack(
                    [
                        lon_i[start : end - 1],
                        lat_i[start : end - 1],
                        lon_i[start + 1 : end],
                        lat_i[start + 1 : end],
                    ]
                ).reshape(-1, 2, 2)
                lc = LineCollection(segments, cmap=cmap, norm=norm, *args, **kwargs)
                lc.set_array(
                    # color of a segment is the average of its two data points
                    np.convolve(colors_i[start:end], [0.5, 0.5], mode="valid")
                )
                ax.add_collection(lc)

            start = end

    # set axis limits
    ax.set_xlim([np.min(longitude), np.max(longitude)])
    ax.set_ylim([np.min(latitude), np.max(latitude)])

    return cb



import unittest

import clouddrift.raggedarray as ragged
from clouddrift.adapters import utils


class DisableProgressTestCase(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        utils._DISABLE_SHOW_PROGRESS = True
        ragged._DISABLE_SHOW_PROGRESS = True

    def tearDown(self) -> None:
        super().tearDown()
        utils._DISABLE_SHOW_PROGRESS = False
        ragged._DISABLE_SHOW_PROGRESS = False



import logging
import os

logging.basicConfig(
    level=os.environ.get("LOG_LEVEL", logging.getLevelName(logging.INFO)).upper(),
    format="[%(asctime)s][%(levelname)s][%(module)s][%(funcName)s][Ln %(lineno)s] - %(message)s",
)



"""
This module provides functions to easily access ragged array datasets. If the datasets are
not accessed via cloud storage platforms or are not found on the local filesystem,
they will be downloaded from their upstream repositories and stored for later access
(~/.clouddrift for UNIX-based systems).
"""

import os
from collections.abc import Callable

import xarray as xr

from clouddrift import adapters
from clouddrift.adapters.hurdat2 import _BasinOption


def gdp1h(decode_times: bool = True) -> xr.Dataset:
    """Returns the latest version of the NOAA Global Drifter Program (GDP) hourly
    dataset as a ragged array Xarray dataset.

    The data is accessed from zarr archive hosted on a public AWS S3 bucket accessible at
    https://registry.opendata.aws/noaa-oar-hourly-gdp/. Original data source from NOAA NCEI
    is https://doi.org/10.25921/x46c-3620).

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        Hourly GDP dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import gdp1h
    >>> ds = gdp1h()
    >>> ds
    <xarray.Dataset>
    Dimensions:                (traj: 19396, obs: 197214787)
    Coordinates:
        id                     (traj) int64 ...
        time                   (obs) datetime64[ns] ...
    Dimensions without coordinates: traj, obs
    Data variables: (12/60)
        BuoyTypeManufacturer   (traj) |S20 ...
        BuoyTypeSensorArray    (traj) |S20 ...
        CurrentProgram         (traj) float32 ...
        DeployingCountry       (traj) |S20 ...
        DeployingShip          (traj) |S20 ...
        DeploymentComments     (traj) |S20 ...
        ...                     ...
        start_lat              (traj) float32 ...
        start_lon              (traj) float32 ...
        typebuoy               (traj) |S10 ...
        typedeath              (traj) int8 ...
        ve                     (obs) float32 ...
        vn                     (obs) float32 ...
    Attributes: (12/16)
        Conventions:       CF-1.6
        acknowledgement:   Elipot, Shane; Sykulski, Adam; Lumpkin, Rick; Centurio...
        contributor_name:  NOAA Global Drifter Program
        contributor_role:  Data Acquisition Center
        date_created:      2023-09-08T17:05:12.130123
        doi:               10.25921/x46c-3620
        ...                ...
        processing_level:  Level 2 QC by GDP drifter DAC
        publisher_email:   aoml.dftr@noaa.gov
        publisher_name:    GDP Drifter DAC
        publisher_url:     https://www.aoml.noaa.gov/phod/gdp
        summary:           Global Drifter Program hourly data
        title:             Global Drifter Program hourly drifting buoy collection

    See Also
    --------
    :func:`gdp6h`
    """
    url = "https://noaa-oar-hourly-gdp-pds.s3.amazonaws.com/latest/gdp-v2.01.1.zarr"
    ds = xr.open_dataset(url, engine="zarr", decode_times=decode_times)
    return ds


def gdp6h(decode_times: bool = True) -> xr.Dataset:
    """Returns the NOAA Global Drifter Program (GDP) 6-hourly dataset as a ragged array
    Xarray dataset.

    The data is accessed from zarr archive hosted on a public AWS S3 bucket accessible at
    s3://noaa-oar-hourly-gdp-pds/experimental/. Original data source from NOAA's Atlantic
    Oceanographic and Meteorological Laboratory (AOML) accessible at
    https://www.aoml.noaa.gov/phod/gdp/index.php.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        6-hourly GDP dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import gdp6h
    >>> ds = gdp6h()
    >>> ds
    <xarray.Dataset> Size: 2GB
    Dimensions:                (traj: 27647, obs: 46535470)
    Coordinates:
    id                     (traj) int64 221kB ...
    time                   (obs) datetime64[ns] 372MB ...
    Dimensions without coordinates: traj, obs
    Data variables: (12/49)
    BuoyTypeManufacturer   (traj) |S20 553kB ...
    BuoyTypeSensorArray    (traj) |S20 553kB ...
    CurrentProgram         (traj) float64 221kB ...
    DeployingCountry       (traj) |S20 553kB ...
    DeployingShip          (traj) |S20 553kB ...
    DeploymentComments     (traj) |S20 553kB ...
    ...                     ...
    start_lon              (traj) float32 111kB ...
    temp                   (obs) float32 186MB ...
    typebuoy               (traj) |S10 276kB ...
    typedeath              (traj) int8 28kB ...
    ve                     (obs) float32 186MB ...
    vn                     (obs) float32 186MB ...
    Attributes: (12/18)
    Conventions:          CF-1.6
    acknowledgement:      Lumpkin, Rick; Centurioni, Luca (2019). NOAA Global...
    contributor_name:     NOAA Global Drifter Program
    contributor_role:     Data Acquisition Center
    date_created:         2024-04-04T13:44:01.176967
    doi:                  10.25921/7ntx-z961
    ...                   ...
    publisher_name:       GDP Drifter DAC
    publisher_url:        https://www.aoml.noaa.gov/phod/gdp
    summary:              Global Drifter Program six-hourly data
    time_coverage_end:    2023-10-18:18:00:00Z
    time_coverage_start:  1979-02-15:00:00:00Z
    title:                Global Drifter Program drifting buoy collection

    See Also
    --------
    :func:`gdp1h`
    """
    url = "https://noaa-oar-hourly-gdp-pds.s3.amazonaws.com/experimental/gdp6h_ragged_sep23.zarr"
    ds = xr.open_dataset(url, decode_times=decode_times, engine="zarr")
    return ds


def gdp_source(
    tmp_path: str = adapters.gdp_source._TMP_PATH,
    max: int | None = None,
    skip_download: bool = False,
    use_fill_values: bool = True,
    decode_times: bool = True,
) -> xr.Dataset:
    """Returns the NOAA Global Drifter Program (GDP) source (raw) dataset as a ragged array
    Xarray dataset.

    The function will first look for the ragged-array dataset on the local
    filesystem. If it is not found, the dataset will be downloaded using the
    corresponding adapter function and stored as zarr archive for later access.

    The data is accessed from a public HTTPS server at NOAA's Atlantic
    Oceanographic and Meteorological Laboratory (AOML) at
    https://www.aoml.noaa.gov/ftp/pub/phod/pub/pazos/data/shane/sst/.

    Parameters
    ----------
    tmp_path: str, default adapter temp path (default)
        Temporary path where intermediary files are stored.
    max: int, optional
        Maximum number of files to retrieve and parse to generate the aggregate file. Mainly used
        for testing purposes.
    skip_download: bool, False (default)
        If True, skips downloading the data files and the code assumes the files have already been downloaded.
        This is mainly used to skip downloading files if the remote doesn't provide the HTTP  Last-Modified header.
    use_fill_values: bool, True (default)
        When True, missing metadata fields are replaced with fill values. When False and no metadata
        is found for a given drifter its observations are ignored.
    decode_times : bool, True (default)
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        source GDP dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import gdp_source
    >>> ds = gdp_source()
    >>> ds
    <xarray.Dataset> Size: ...
    Dimensions:            (traj: ..., obs: ...)
    Coordinates:
        id                 (traj) int64 222kB ...
        obs_index          (obs) int32 1GB ...
    Dimensions without coordinates: traj, obs
    Data variables: (12/22)
        buoys_type         (traj) <U5 ...
        death_code         (traj) int64 ...
        drogue             (obs) float32 ...
        drogue_off_date    (traj) datetime64[ns] ...
        end_date           (traj) datetime64[ns] ...
        end_lat            (traj) float64 222kB ...
        ...                 ...
        sst                (obs) float32 ...
        start_date         (traj) datetime64[ns] ...
        start_lat          (traj) float64 ...
        start_lon          (traj) float64 ...
        voltage            (obs) float32 ...
        wmo_number         (traj) int64 ...
    """
    file_selection_label = "all" if max is None else f"first-{max}"
    return _dataset_filecache(
        f"gdpsource_agg_{file_selection_label}.zarr",
        decode_times,
        lambda: adapters.gdp_source.to_raggedarray(
            tmp_path, skip_download, max, use_fill_values=use_fill_values
        ),
    )


def glad(decode_times: bool = True) -> xr.Dataset:
    """Returns the Grand LAgrangian Deployment (GLAD) dataset as a ragged array
    Xarray dataset.

    The function will first look for the ragged-array dataset on the local
    filesystem. If it is not found, the dataset will be downloaded using the
    corresponding adapter function and stored for later access.

    The upstream data is available at https://doi.org/10.7266/N7VD6WC8.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        GLAD dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import glad
    >>> ds = glad()
    >>> ds
    <xarray.Dataset>
    Dimensions:         (obs: 1602883, traj: 297)
    Coordinates:
      time            (obs) datetime64[ns] ...
      id              (traj) object ...
    Data variables:
      latitude        (obs) float32 ...
      longitude       (obs) float32 ...
      position_error  (obs) float32 ...
      u               (obs) float32 ...
      v               (obs) float32 ...
      velocity_error  (obs) float32 ...
      rowsize         (traj) int64 ...
    Attributes:
      title:        GLAD experiment CODE-style drifter trajectories (low-pass f...
      institution:  Consortium for Advanced Research on Transport of Hydrocarbo...
      source:       CODE-style drifters
      history:      Downloaded from https://data.gulfresearchinitiative.org/dat...
      references:   Özgökmen, Tamay. 2013. GLAD experiment CODE-style drifter t...

    Reference
    ---------
    Özgökmen, Tamay. 2013. GLAD experiment CODE-style drifter trajectories (low-pass filtered, 15 minute interval records), northern Gulf of Mexico near DeSoto Canyon, July-October 2012. Distributed by: Gulf of Mexico Research Initiative Information and Data Cooperative (GRIIDC), Harte Research Institute, Texas A&M University–Corpus Christi. doi:10.7266/N7VD6WC8
    """
    return _dataset_filecache("glad.nc", decode_times, adapters.glad.to_xarray)


def hurdat2(
    basin: _BasinOption = "both",
    decode_times: bool = True,
    tmp_path: str = adapters.hurdat2._DEFAULT_FILE_PATH,
) -> xr.DataArray:
    """Returns the revised hurricane database (HURDAT2) as a ragged array xarray dataset.

    The function will first look for the ragged array dataset on the local
    filesystem. If it is not found, the dataset will be downloaded using the
    corresponding adapter function and stored for later access.

    The upstream data is available at https://www.aoml.noaa.gov/hrd/hurdat/hurdat2.html.

    Parameters
    ----------
    basin : "atlantic", "pacific", "both" (default)
        Specify which ocean basin to download storm data for. Current available options are the
        Atlantic Ocean "atlantic", Pacific Ocean "pacific" and, both "both" to download
        storm data for both oceans.
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        HURDAT2 dataset as a ragged array.

    Standard usage of the dataset.

    >>> from clouddrift.datasets import hurdat2
    >>> ds = hurdat2()
    >>> ds
    <xarray.Dataset>
    Dimensions:                          (traj: ..., obs: ...)
    Coordinates:
        atcf_identifier                  (traj) <U8 ...
        time                             (obs) datetime64[ns] ...
    Dimensions without coordinates: traj, obs
    Data variables: (12/23)
        basin                            (traj) <U2 ...
        year                             (traj) int64 ...
        rowsize                          (traj) int64 ...
        record_identifier                (obs) <U1 ...
        system_status                    (obs) <U2 ...
        ...                               ...
        max_med_wind_radius_nw           (obs) float64 ...
        max_high_wind_radius_ne          (obs) float64 ...
        max_high_wind_radius_se          (obs) float64 ...
        max_high_wind_radius_sw          (obs) float64 ...
        max_high_wind_radius_nw          (obs) float64 ...
        max_sustained_wind_speed_radius  (obs) float64 ...
    Attributes:
        title:            HURricane DATa 2nd generation (HURDAT2)
        date_created:     ...
        publisher_name:   NOAA AOML Hurricane Research Division
        publisher_email:  AOML.HRDWebmaster@noaa.gov
        publisher_url:    https://www.aoml.noaa.gov/hrd/hurdat/Data_Storm.html
        institution:      NOAA Atlantic Oceanographic and Meteorological Laboratory
        summary:          The National Hurricane Center (NHC) conducts a post-sto...
    ...

    To only retrieve records for the Atlantic Ocean basin.

    >>> from clouddrift.datasets import hurdat2
    >>> ds = hurdat2(basin="atlantic")
    >>> ds
    <xarray.Dataset>
    ...

    Reference
    ---------
    https://www.aoml.noaa.gov/hrd/hurdat/Data_Storm.html.
    """
    return _dataset_filecache(
        f"hurdat2_{basin}.nc",
        decode_times,
        lambda: adapters.hurdat2.to_raggedarray(basin, tmp_path).to_xarray(),
    )


def mosaic(decode_times: bool = True) -> xr.Dataset:
    """Returns the MOSAiC sea-ice drift dataset as a ragged array Xarray dataset.

    The function will first look for the ragged-array dataset on the local
    filesystem. If it is not found, the dataset will be downloaded using the
    corresponding adapter function and stored for later access.

    The upstream data is available at https://arcticdata.io/catalog/view/doi:10.18739/A2KP7TS83.

    Reference
    ---------
    Angela Bliss, Jennifer Hutchings, Philip Anderson, Philipp Anhaus,
    Hans Jakob Belter, Jørgen Berge, Vladimir Bessonov, Bin Cheng, Sylvia Cole,
    Dave Costa, Finlo Cottier, Christopher J Cox, Pedro R De La Torre, Dmitry V Divine,
    Gilbert Emzivat, Ying-Chih Fang, Steven Fons, Michael Gallagher, Maxime Geoffrey,
    Mats A Granskog, ... Guangyu Zuo. (2022). Sea ice drift tracks from the Distributed
    Network of autonomous buoys deployed during the Multidisciplinary drifting Observatory
    for the Study of Arctic Climate (MOSAiC) expedition 2019 - 2021. Arctic Data Center.
    doi:10.18739/A2KP7TS83.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        MOSAiC sea-ice drift dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import mosaic
    >>> ds = mosaic()
    >>> ds
    <xarray.Dataset>
    Dimensions:                     (obs: 1926226, traj: 216)
    Coordinates:
        time                        (obs) datetime64[ns] ...
        id                          (traj) object ...
    Dimensions without coordinates: obs, traj
    Data variables: (12/19)
        latitude                    (obs) float64 ...
        longitude                   (obs) float64 ...
        Deployment Leg              (traj) int64 ...
        DN Station ID               (traj) object ...
        IMEI                        (traj) object ...
        Deployment Date             (traj) datetime64[ns] ...
        ...                          ...
        Buoy Type                   (traj) object ...
        Manufacturer                (traj) object ...
        Model                       (traj) object ...
        PI                          (traj) object ...
        Data Authors                (traj) object ...
        rowsize                     (traj) int64 ...
    """
    return _dataset_filecache("mosaic.nc", decode_times, adapters.mosaic.to_xarray)


def spotters(decode_times: bool = True) -> xr.Dataset:
    """Returns the Sofar Ocean Spotter drifters ragged array dataset as an Xarray dataset.

    The data is accessed from a zarr archive hosted on a public AWS S3 bucket accessible
    at https://sofar-spotter-archive.s3.amazonaws.com/spotter_data_bulk_zarr.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        Sofar ocean floats dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import spotters
    >>> ds = spotters()
    >>> ds
    <xarray.Dataset>
    Dimensions:                (index: 6390651, trajectory: 871)
    Coordinates:
        time                   (index) datetime64[ns] ...
      * trajectory             (trajectory) object 'SPOT-010001' ... 'SPOT-1975'
    Dimensions without coordinates: index
    Data variables:
        latitude               (index) float64 ...
        longitude              (index) float64 ...
        meanDirection          (index) float64 ...
        meanDirectionalSpread  (index) float64 ...
        meanPeriod             (index) float64 ...
        peakDirection          (index) float64 ...
        peakDirectionalSpread  (index) float64 ...
        peakPeriod             (index) float64 ...
        rowsize                (trajectory) int64 ...
        significantWaveHeight  (index) float64 ...
    Attributes:
        author:         Isabel A. Houghton
        creation_date:  2023-10-18 00:43:55.333537
        email:          isabel.houghton@sofarocean.com
        institution:    Sofar Ocean
        references:     https://content.sofarocean.com/hubfs/Spotter%20product%20...
        source:         Spotter wave buoy
        title:          Sofar Spotter Data Archive - Bulk Wave Parameters
    """
    url = "https://sofar-spotter-archive.s3.amazonaws.com/spotter_data_bulk_zarr"
    return xr.open_dataset(url, engine="zarr", decode_times=decode_times)


def subsurface_floats(decode_times: bool = True) -> xr.Dataset:
    """Returns the subsurface floats dataset as a ragged array Xarray dataset.

    The data is accessed from a public HTTPS server at NOAA's Atlantic
    Oceanographic and Meteorological Laboratory (AOML) accessible at
    https://www.aoml.noaa.gov/phod/gdp/index.php.

    The upstream data is available at
    https://www.aoml.noaa.gov/phod/float_traj/files/allFloats_12122017.mat.

    This dataset of subsurface float observations was compiled by the WOCE Subsurface
    Float Data Assembly Center (WFDAC) in Woods Hole maintained by Andree Ramsey and
    Heather Furey and copied to NOAA/AOML in October 2014 (version 1) and in December
    2017 (version 2). Subsequent updates will be included as additional appropriate
    float data, quality controlled by the appropriate principal investigators, is
    submitted for inclusion.

    Note that these observations are collected by ALACE/RAFOS/Eurofloat-style
    acoustically-tracked, neutrally-buoyant subsurface floats which collect data while
    drifting beneath the ocean surface. These data are the result of the effort and
    resources of many individuals and institutions. You are encouraged to acknowledge
    the work of the data originators and Data Centers in publications arising from use
    of these data.

    The float data were originally divided by project at the WFDAC. Here they have been
    compiled in a single Matlab data set. See here for more information on the variables
    contained in these files.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        Subsurface floats dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import subsurface_floats
    >>> ds = subsurface_floats()
    >>> ds
    <xarray.Dataset>
    Dimensions:   (traj: 2193, obs: 1402840)
    Coordinates:
        id        (traj) uint16 ...
        time      (obs) datetime64[ns] ...
    Dimensions without coordinates: traj, obs
    Data variables: (12/13)
        expList   (traj) object ...
        expName   (traj) object ...
        expOrg    (traj) object ...
        expPI     (traj) object ...
        indexExp  (traj) uint8 ...
        fltType   (traj) object ...
        ...        ...
        lon       (obs) float64 ...
        lat       (obs) float64 ...
        pres      (obs) float64 ...
        temp      (obs) float64 ...
        ve        (obs) float64 ...
        vn        (obs) float64 ...
    Attributes:
        title:            Subsurface float trajectories dataset
        history:          December 2017 (version 2)
        date_created:     2023-11-14T22:30:38.831656
        publisher_name:   WOCE Subsurface Float Data Assembly Center and NOAA AOML
        publisher_url:    https://www.aoml.noaa.gov/phod/float_traj/data.php
        license:          freely available
        acknowledgement:  Maintained by Andree Ramsey and Heather Furey from the ...

    References
    ----------
    WOCE Subsurface Float Data Assembly Center (WFDAC) https://www.aoml.noaa.gov/phod/float_traj/index.php
    """
    return _dataset_filecache(
        "subsurface_floats.nc", decode_times, adapters.subsurface_floats.to_xarray
    )


def yomaha(decode_times: bool = True) -> xr.Dataset:
    """Returns the YoMaHa dataset as a ragged array Xarray dataset.

    The function will first look for the ragged-array dataset on the local
    filesystem. If it is not found, the dataset will be downloaded using the
    corresponding adapter function and stored for later access. The upstream
    data is available at http://apdrc.soest.hawaii.edu/projects/yomaha/.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        YoMaHa'07 dataset as a ragged array

    Examples
    --------

    >>> from clouddrift.datasets import yomaha
    >>> ds = yomaha()
    >>> ds
    <xarray.Dataset>
    Dimensions:     (obs: 1926743, traj: 12196)
    Coordinates:
        time_d      (obs) datetime64[ns] ...
        time_s      (obs) datetime64[ns] ...
        time_lp     (obs) datetime64[ns] ...
        time_lc     (obs) datetime64[ns] ...
        id          (traj) int64 ...
    Dimensions without coordinates: obs, traj
    Data variables: (12/27)
        lon_d       (obs) float64 ...
        lat_d       (obs) float64 ...
        pres_d      (obs) float32 ...
        ve_d        (obs) float32 ...
        vn_d        (obs) float32 ...
        err_ve_d    (obs) float32 ...
        ...          ...
        cycle       (obs) int64 ...
        time_inv    (obs) int64 ...
        rowsize     (traj) int64 ...
        wmo_id      (traj) int64 ...
        dac_id      (traj) int64 ...
        float_type  (traj) int64 ...
    Attributes:
        title:           YoMaHa'07: Velocity data assessed from trajectories of A...
        history:         Dataset updated on Tue Jun 28 03:14:34 HST 2022
        date_created:    2023-12-08T00:52:08.478075
        publisher_name:  Asia-Pacific Data Research Center
        publisher_url:   http://apdrc.soest.hawaii.edu/index.php
        license:         Creative Commons Attribution 4.0 International License..

    Reference
    ---------
    Lebedev, K. V., Yoshinari, H., Maximenko, N. A., & Hacker, P. W. (2007). Velocity data
    assessed  from trajectories of Argo floats at parking level and at the sea
    surface. IPRC Technical Note, 4(2), 1-16.
    """
    return _dataset_filecache("yomaha.nc", decode_times, adapters.yomaha.to_xarray)


def andro(decode_times: bool = True) -> xr.Dataset:
    """Returns the ANDRO as a ragged array Xarray dataset.

    The function will first look for the ragged-array dataset on the local
    filesystem. If it is not found, the dataset will be downloaded using the
    corresponding adapter function and stored for later access. The upstream
    data is available at https://www.seanoe.org/data/00360/47077/.

    Parameters
    ----------
    decode_times : bool, optional
        If True, decode the time coordinate into a datetime object. If False, the time
        coordinate will be an int64 or float64 array of increments since the origin
        time indicated in the units attribute. Default is True.

    Returns
    -------
    xarray.Dataset
        ANDRO dataset as a ragged array
    Examples
    --------
    >>> from clouddrift.datasets import andro
    >>> ds = andro()
    >>> ds
    <xarray.Dataset> Size: 110MB
    Dimensions:     (obs: 510630, traj: 3344)
    Coordinates:
        time_d      (obs) datetime64[ns] 4MB ...
        time_s      (obs) datetime64[ns] 4MB ...
        time_lp     (obs) datetime64[ns] 4MB ...
        time_lc     (obs) datetime64[ns] 4MB ...
        id          (traj) float32 13kB ...
    Dimensions without coordinates: obs, traj
    Data variables: (12/33)
        lon_d       (obs) float64 4MB ...
        lat_d       (obs) float64 4MB ...
        pres_d      (obs) float32 2MB ...
        temp_d      (obs) float32 2MB ...
        sal_d       (obs) float32 2MB ...
        ve_d        (obs) float32 2MB ...
        ...          ...
        lon_lc      (obs) float64 4MB ...
        lat_lc      (obs) float64 4MB ...
        surf_fix    (obs) float32 2MB ...
        cycle       (obs) float32 2MB ...
        profile_id  (obs) float32 2MB ...
        rowsize     (traj) int64 27kB ...
    Attributes:
        title:           ANDRO: An Argo-based deep displacement dataset (Quality ...
        history:         Dataset updated on 2024-03-25
        date_created:    2024-07-02T12:11:50.643492
        publisher_name:  SEANOE (SEA scieNtific Open data Edition)
        publisher_url:   https://www.seanoe.org/data/00360/47077/
        license:         Creative Commons Attribution 4.0 International License (...

    Reference
    ---------
    Ollitrault Michel, Rannou Philippe, Brion Emilie, Cabanes Cecile, Piron Anne, Reverdin Gilles,
    Kolodziejczyk Nicolas (2022). ANDRO: An Argo-based deep displacement dataset.
    SEANOE. https://doi.org/10.17882/47077
    """
    return _dataset_filecache("andro.nc", decode_times, adapters.andro.to_xarray)


def _dataset_filecache(
    filename: str, decode_times: bool, get_ds: Callable[[], xr.Dataset]
):
    os.path
    _, ext = os.path.splitext(filename)
    clouddrift_path = (
        os.path.expanduser("~/.clouddrift")
        if not os.getenv("CLOUDDRIFT_PATH")
        else os.getenv("CLOUDDRIFT_PATH")
    )
    fp = f"{clouddrift_path}/data/{filename}"
    os.makedirs(os.path.dirname(fp), exist_ok=True)

    if not os.path.exists(fp):
        ds = get_ds()
        if ext == ".nc":
            ds.to_netcdf(fp)
        elif ext == ".zarr":
            ds.to_zarr(fp)
        else:
            ds.to_netcdf(fp)

    engine = None
    if ext == ".zarr":
        engine = "zarr"

    return xr.open_dataset(fp, decode_times=decode_times, engine=engine)



"""
This module provides adapters to custom datasets.
Each adapter module provides convenience functions and metadata to convert a
custom dataset to a `clouddrift.RaggedArray` instance.
Currently, clouddrift provides adapter modules for the hourly Global
Drifter Program (GDP) dataset, the 6-hourly GDP dataset, 15-minute Grand LAgrangian
Deployment (GLAD) dataset, and the MOSAiC sea-ice drift dataset. More adapters will be added
in the future.
"""

import clouddrift.adapters.andro as andro
import clouddrift.adapters.gdp.gdp1h as gdp1h
import clouddrift.adapters.gdp.gdp6h as gdp6h
import clouddrift.adapters.gdp.gdpsource as gdp_source
import clouddrift.adapters.glad as glad
import clouddrift.adapters.hurdat2 as hurdat2
import clouddrift.adapters.mosaic as mosaic
import clouddrift.adapters.subsurface_floats as subsurface_floats
import clouddrift.adapters.utils as utils
import clouddrift.adapters.yomaha as yomaha

__all__ = [
    "andro",
    "gdp1h",
    "hurdat2",
    "gdp6h",
    "glad",
    "mosaic",
    "subsurface_floats",
    "yomaha",
    "gdp_source",
    "utils",
]



from __future__ import annotations

import asyncio
import datetime
import logging
import os
import tempfile
import warnings
from collections import defaultdict
from concurrent.futures import Future, ProcessPoolExecutor, as_completed
from typing import Callable

import numpy as np
import pandas as pd
import xarray as xr
from tqdm.asyncio import tqdm

from clouddrift.adapters.gdp import get_gdp_metadata
from clouddrift.adapters.utils import download_with_progress
from clouddrift.ragged import subset
from clouddrift.raggedarray import RaggedArray

_DATA_URL = "https://www.aoml.noaa.gov/ftp/pub/phod/pub/pazos/data/shane/sst"
_TMP_PATH = os.path.join(tempfile.gettempdir(), "clouddrift", "gdpsource")
_FILENAME_TEMPLATE = "buoydata_{start}_{end}_{suffix}.dat.gz"
_SECONDS_IN_DAY = 86_400

_COORDS = ["id", "obs_index"]

_DATA_VARS = [
    "latitude",
    "longitude",
    "position_datetime",
    "sensor_datetime",
    "drogue",
    "sst",
    "voltage",
    "sensor4",
    "sensor5",
    "sensor6",
    "qualityIndex",
]

_METADATA_VARS = [
    "rowsize",
    "wmo_number",
    "program_number",
    "buoys_type",
    "start_date",
    "start_lat",
    "start_lon",
    "end_date",
    "end_lat",
    "end_lon",
    "drogue_off_date",
    "death_code",
]

_VARS_FILL_MAP: dict = {
    "wmo_number": -999,
    "program_number": -999,
    "buoys_type": "N/A",
    "start_date": np.datetime64("1970-01-01 00:00:00"),
    "start_lat": -999,
    "start_lon": -999,
    "end_date": np.datetime64("1970-01-01 00:00:00"),
    "end_lat": -999,
    "end_lon": -999,
    "drogue_off_date": np.datetime64("1970-01-01 00:00:00"),
    "death_code": -999,
}

_VAR_DTYPES: dict = {
    "rowsize": np.int64,
    "wmo_number": np.int64,
    "program_number": np.int64,
    "buoys_type": np.str_,
    "start_date": np.dtype("datetime64[ns]"),
    "start_lat": np.float64,
    "start_lon": np.float64,
    "end_date": np.dtype("datetime64[ns]"),
    "end_lat": np.float64,
    "end_lon": np.float64,
    "drogue_off_date": np.dtype("datetime64[ns]"),
    "death_code": np.int64,
}


_INPUT_COLS = [
    "id",
    "posObsMonth",
    "posObsDay",
    "posObsYear",
    "latitude",
    "longitude",
    "senObsMonth",
    "senObsDay",
    "senObsYear",
    "drogue",
    "sst",
    "voltage",
    "sensor4",
    "sensor5",
    "sensor6",
    "qualityIndex",
]


_INPUT_COLS_DTYPES = {
    "id": np.int64,
    "posObsMonth": np.int8,
    "posObsDay": np.float64,
    "posObsYear": np.int16,
    "latitude": np.float32,
    "longitude": np.float32,
    "qualityIndex": np.float32,
    "senObsMonth": np.int8,
    "senObsDay": np.float64,
    "senObsYear": np.int16,
    "drogue": np.float32,
    "sst": np.float32,
    "voltage": np.float32,
    "sensor4": np.float32,
    "sensor5": np.float32,
    "sensor6": np.float32,
}


VARS_ATTRS: dict = {
    "id": {"long_name": "Global Drifter Program Buoy ID", "units": "-"},
    "rowsize": {
        "long_name": "Number of observations per trajectory",
        "sample_dimension": "obs",
        "units": "-",
    },
    "wmo_number": {
        "long_name": "World Meteorological Organization buoy identification number",
        "units": "-",
    },
    "program_number": {
        "long_name": "Current Program",
        "units": "-",
    },
    "buoys_type": {
        "long_name": "Buoy type (see https://www.aoml.noaa.gov/phod/dac/dirall.html)",
        "units": "-",
    },
    "start_date": {
        "long_name": "First good date and time derived by DAC quality control",
    },
    "start_lon": {
        "long_name": "First good longitude derived by DAC quality control",
        "units": "degrees_east",
    },
    "start_lat": {
        "long_name": "First good latitude derived by DAC quality control",
        "units": "degrees_north",
    },
    "end_date": {
        "long_name": "Last good date and time derived by DAC quality control",
    },
    "end_lon": {
        "long_name": "Last good longitude derived by DAC quality control",
        "units": "degrees_east",
    },
    "end_lat": {
        "long_name": "Last good latitude derived by DAC quality control",
        "units": "degrees_north",
    },
    "drogue_off_date": {
        "long_name": "Date and time of drogue loss",
    },
    "death_code": {
        "long_name": "Type of death",
        "units": "-",
        "comments": "0 (buoy still alive), 1 (buoy ran aground), 2 (picked up by vessel), 3 (stop transmitting), 4 (sporadic transmissions), 5 (bad batteries), 6 (inactive status)",
    },
    "position_datetime": {
        "comments": "Position datetime derived from the year, month, day and time (represented as a ratio of a day) columns found in the source dataset that represent when the position of the drifter was measured. This value is only different from the sensor_datetime when the position of the drifter was determined onboard the Argos satellites using the doppler shift.",
    },
    "sensor_datetime": {
        "comments": "Sensor datetime derived from the year, month, day and time (represented as a ratio of a day) columns found in the source dataset that represent when the sensor (like temp) data is recorded",
    },
    "longitude": {"long_name": "Longitude", "units": "degrees_east"},
    "latitude": {"long_name": "Latitude", "units": "degrees_north"},
    "drogue": {
        "long_name": "Drogue",
        "units": "-",
        "comments": "Values returned by drogue sensor",
    },
    "sst": {
        "long_name": "sea water temperature",
        "units": "degree_Celsius",
        "comments": "sea water temperature from drifting buoy measurements",
    },
    "voltage": {
        "long_name": "Voltage",
        "units": "V",
    },
    "sensor4": {
        "long_name": "Sensor 4",
        "units": "-",
        "comments": "placeholders for additional sensors such as barometers, salinity etc.",
    },
    "sensor5": {
        "long_name": "Sensor 5",
        "units": "-",
        "comments": "placeholders for additional sensors such as barometers, salinity etc.",
    },
    "sensor6": {
        "long_name": "Sensor 6",
        "units": "-",
        "comments": "placeholders for additional sensors such as barometers, salinity etc.",
    },
    "qualityIndex": {
        "long_name": "Quality Index",
        "units": "-",
    },
}

ATTRS = {
    "title": "Global Drifter Program source drifter dataset",
    "Conventions": "CF-1.6",
    "date_created": datetime.datetime.now().isoformat(),
    "publisher_name": "GDP Drifter DAC",
    "publisher_email": "aoml.dftr@noaa.gov",
    "publisher_url": "https://www.aoml.noaa.gov/phod/gdp",
    "license": "freely available",
    "processing_level": "source files",
    "metadata_link": "https://www.aoml.noaa.gov/phod/dac/dirall.html",
    "contributor_name": "NOAA Global Drifter Program",
    "contributor_role": "Data Acquisition Center",
    "institution": "NOAA Atlantic Oceanographic and Meteorological Laboratory",
    "summary": "Global Drifter Program source (raw) data",
}

_logger = logging.getLogger(__name__)


def _get_download_list(tmp_path: str) -> list[tuple[str, str]]:
    suffix = "rawfiles"
    batches = [(1, 5000), (5001, 10_000), (10_001, 15_000), (15_001, "current")]

    requests = list()

    for start, end in batches:
        filename = _FILENAME_TEMPLATE.format(start=start, end=end, suffix=suffix)
        requests.append((f"{_DATA_URL}/{filename}", os.path.join(tmp_path, filename)))
    return requests


def _rowsize(id_, **kwargs) -> int:
    df: pd.DataFrame | None = kwargs.get("data_df")

    if df is None:
        raise KeyError(
            "Missing `data_df`, please pass them into the `from_files` method"
        )

    traj_data_df = df[df["id"] == id_]
    return len(traj_data_df)


def _preprocess(id_, **kwargs) -> xr.Dataset:
    md_df: pd.DataFrame | None = kwargs.get("md_df")
    data_df: pd.DataFrame | None = kwargs.get("data_df")
    use_fill_values: bool = kwargs.get("use_fill_values", False)

    if md_df is None or data_df is None:
        raise KeyError(
            "Missing `md_df` or `data_df`, pass them into the `from_files` method"
        )

    traj_md_df = md_df[md_df["ID"] == id_]
    traj_data_df = data_df[data_df["id"] == id_]
    rowsize = len(traj_data_df)

    md_variables = {
        "rowsize": np.array([rowsize], dtype=np.int64),
    }

    if use_fill_values and len(traj_md_df) == 0:
        for md_var in _METADATA_VARS:
            if md_var == "rowsize":
                continue

            fill_val = _VARS_FILL_MAP.get(md_var)
            if fill_val is None:
                raise ValueError(
                    f"Fill value missing for metadata variable: `{md_var}`"
                )

            md_variables.update({md_var: np.array([fill_val])})
    else:
        md_variables.update(
            {
                "wmo_number": traj_md_df[["WMO_number"]].values.flatten(),
                "program_number": traj_md_df[["program_number"]].values.flatten(),
                "buoys_type": traj_md_df[["buoys_type"]].values.flatten(),
                "start_date": traj_md_df[["Start_date"]].values.flatten(),
                "start_lat": traj_md_df[["Start_lat"]].values.flatten(),
                "start_lon": traj_md_df[["Start_lon"]].values.flatten(),
                "end_date": traj_md_df[["End_date"]].values.flatten(),
                "end_lat": traj_md_df[["End_lat"]].values.flatten(),
                "end_lon": traj_md_df[["End_lon"]].values.flatten(),
                "drogue_off_date": traj_md_df[["Drogue_off_date"]].values.flatten(),
                "death_code": traj_md_df[["death_code"]].values.flatten(),
            }
        )

    for md_var in _METADATA_VARS:
        var = md_variables.get(md_var)

        if var is None:
            raise ValueError(f"Metadata variable `{md_var}` cannot be found")

        md_variables.update({md_var: var.astype(_VAR_DTYPES.get(md_var))})

    variables = {k: (["traj"], md_variables[k]) for k in md_variables}
    data_vars = {
        var: (["obs"], traj_data_df[[var]].values.flatten()) for var in _DATA_VARS
    }

    coords = {
        "id": (["traj"], np.array([id_]).astype(np.int64)),
        "obs_index": (
            ["obs"],
            traj_data_df[["obs_index"]].values.flatten().astype(np.int32),
        ),
    }

    variables.update(data_vars)
    dataset = xr.Dataset(variables, coords=coords)
    return dataset


def _apply_remove(df: pd.DataFrame, filters: list[Callable]) -> pd.DataFrame:
    temp_df = df
    for filter_ in filters:
        mask = filter_(temp_df)
        temp_df = temp_df[~mask]
    return temp_df


def _apply_transform(
    df: pd.DataFrame,
    transforms: dict[str, tuple[list[str], Callable]],
) -> pd.DataFrame:
    tmp_df = df
    for output_col in transforms.keys():
        input_cols, func = transforms[output_col]
        args = list()
        for col in input_cols:
            arg = df[[col]].values.flatten()
            args.append(arg)
        tmp_df = tmp_df.assign(**{output_col: func(*args)})
        tmp_df = tmp_df.drop(input_cols, axis=1)
    return tmp_df


def _parse_datetime_with_day_ratio(
    month_series: np.ndarray, day_series: np.ndarray, year_series: np.ndarray
) -> np.ndarray:
    values = list()
    for month, day_with_ratio, year in zip(month_series, day_series, year_series):
        day = day_with_ratio // 1
        dayratio = day_with_ratio - day
        seconds = dayratio * _SECONDS_IN_DAY
        dt_ns = (
            datetime.datetime(year=int(year), month=int(month), day=int(1))
            + datetime.timedelta(days=int(day), seconds=seconds)
        ).timestamp() * 10**9
        values.append(int(dt_ns))
    return np.array(values).astype("datetime64[ns]")


def _process_chunk(
    df_chunk: pd.DataFrame,
    start_idx: int,
    end_idx: int,
    gdp_metadata_df: pd.DataFrame,
    use_fill_values: bool,
) -> dict[int, xr.Dataset]:
    """Process each dataframe chunk. Return a dictionary mapping each drifter to a unique xarray Dataset."""

    # Transform the initial dataframe filtering out rows with really anomolous values
    # examples include: years in the future, years way in the past before GDP program, etc...
    preremove_df_chunk = df_chunk.assign(obs_index=range(start_idx, end_idx))
    df_chunk = _apply_remove(
        preremove_df_chunk,
        filters=[
            # Filter out year values that are in the future or predating the GDP program
            lambda df: (df["posObsYear"] > datetime.datetime.now().year)
            | (df["posObsYear"] < 0),
            lambda df: (df["senObsYear"] > datetime.datetime.now().year)
            | (df["senObsYear"] < 0),
            # Filter out month values that contain non-numeric characters
            lambda df: df["senObsMonth"].astype(np.str_).str.contains(r"[\D]"),
            lambda df: df["posObsMonth"].astype(np.str_).str.contains(r"[\D]"),
            # Filter out drogue values that cannot be interpret as floating point values.
            # (e.g. - have more than one decimal point)
            lambda df: df["drogue"].astype(np.str_).str.match(r"(\d+[\.]+){2,}"),
        ],
    )

    drifter_ds_map = dict[int, xr.Dataset]()

    preremove_len = len(preremove_df_chunk)
    postremove_len = len(df_chunk)

    if preremove_len != postremove_len:
        warnings.warn(
            f"Filters removed {preremove_len - postremove_len} rows from chunk"
        )

    if postremove_len == 0:
        return drifter_ds_map

    df_chunk = df_chunk.astype(_INPUT_COLS_DTYPES)
    df_chunk = _apply_transform(
        df_chunk,
        {
            "position_datetime": (
                ["posObsMonth", "posObsDay", "posObsYear"],
                _parse_datetime_with_day_ratio,
            ),
            "sensor_datetime": (
                ["senObsMonth", "senObsDay", "senObsYear"],
                _parse_datetime_with_day_ratio,
            ),
        },
    )

    # Find and process drifters found and documented in the drifter metadata.
    ids_in_data = np.unique(df_chunk[["id"]].values)
    ids_with_md = np.intersect1d(ids_in_data, gdp_metadata_df[["ID"]].values)

    if len(ids_with_md) < len(ids_in_data):
        warnings.warn(
            "Chunk has drifter ids not found in the metadata table. "
            + "Using fill values"
            if use_fill_values
            else "Ignoring data observations"
            + f" for missing metadata ids: {np.setdiff1d(ids_in_data, ids_with_md)}."
        )

    ra = RaggedArray.from_files(
        indices=ids_in_data if use_fill_values else ids_with_md,
        preprocess_func=_preprocess,
        rowsize_func=_rowsize,
        name_coords=_COORDS,
        name_meta=_METADATA_VARS,
        name_data=_DATA_VARS,
        name_dims={"traj": "rows", "obs": "obs"},
        md_df=gdp_metadata_df,
        data_df=df_chunk,
        use_fill_values=use_fill_values,
        tqdm=dict(disable=True),
    )
    ds = ra.to_xarray()

    for id_ in ids_with_md:
        id_f_ds = subset(ds, dict(id=id_), row_dim_name="traj")
        drifter_ds_map[id_] = id_f_ds
    return drifter_ds_map


def _combine_chunked_drifter_datasets(datasets: list[xr.Dataset]) -> xr.Dataset:
    """Combines several drifter observations found in separate chunks, ordering them
    by the observations row index.
    """
    traj_dataset = xr.concat(
        datasets, dim="obs", coords="minimal", data_vars=_DATA_VARS, compat="override"
    )

    new_rowsize = sum([ds.rowsize.values[0] for ds in datasets])
    traj_dataset["rowsize"] = xr.DataArray(
        np.array([new_rowsize], dtype=np.int64), coords=traj_dataset["rowsize"].coords
    )

    sort_coord = traj_dataset.coords["obs_index"]
    vals: np.ndarray = sort_coord.data
    sort_coord_dim = sort_coord.dims[-1]
    sort_key = vals.argsort()

    for coord_name in _COORDS:
        coord = traj_dataset.coords[coord_name]
        dim = coord.dims[-1]

        if dim == sort_coord_dim:
            sorted_coord = coord.isel({dim: sort_key})
            traj_dataset.coords[coord_name] = sorted_coord

    for varname in _DATA_VARS:
        var = traj_dataset[varname]
        dim = var.dims[-1]
        sorted_var = var.isel({dim: sort_key})
        traj_dataset[varname] = sorted_var

    return traj_dataset


async def _parallel_get(
    sources: list[str],
    gdp_metadata_df: pd.DataFrame,
    chunk_size: int,
    tmp_path: str,
    use_fill_values: bool,
    max_chunks: int | None,
) -> list[xr.Dataset]:
    """Parallel process dataset in chunks leveraging multiprocessing."""
    max_workers = (os.cpu_count() or 0) // 2
    with ProcessPoolExecutor(max_workers=max_workers) as ppe:
        drifter_chunked_datasets: dict[int, list[xr.Dataset]] = defaultdict(list)
        start_idx = 0
        for fp in tqdm(
            sources,
            desc="Loading files",
            unit="file",
            ncols=80,
            total=len(sources),
            position=0,
        ):
            file_chunks = pd.read_csv(
                fp,
                sep=r"\s+",
                header=None,
                names=_INPUT_COLS,
                engine="c",
                compression="gzip",
                chunksize=chunk_size,
            )

            joblist = list[Future]()
            jobmap = dict[Future, pd.DataFrame]()
            for idx, chunk in enumerate(file_chunks):
                if max_chunks is not None and idx >= max_chunks:
                    break
                ajob = ppe.submit(
                    _process_chunk,
                    chunk,
                    start_idx,
                    start_idx + len(chunk),
                    gdp_metadata_df,
                    use_fill_values,
                )
                start_idx += len(chunk)
                jobmap[ajob] = chunk
                joblist.append(ajob)

            bar = tqdm(
                desc="Processing file chunks",
                unit="chunk",
                ncols=80,
                total=len(joblist),
                position=1,
            )

            for ajob in as_completed(jobmap.keys()):
                if (exc := ajob.exception()) is not None:
                    chunk = jobmap[ajob]
                    _logger.warn(f"bad chunk detected, exception: {ajob.exception()}")
                    raise exc

                job_drifter_ds_map: dict[int, xr.Dataset] = ajob.result()
                for id_ in job_drifter_ds_map.keys():
                    drifter_ds = job_drifter_ds_map[id_]
                    drifter_chunked_datasets[id_].append(drifter_ds)
                bar.update()

        combine_jobmap = dict[Future, int]()
        for id_ in drifter_chunked_datasets.keys():
            datasets = drifter_chunked_datasets[id_]

            combine_job = ppe.submit(_combine_chunked_drifter_datasets, datasets)
            combine_jobmap[combine_job] = id_

        bar.close()
        bar = tqdm(
            desc="merging drifter chunks",
            unit="drifter",
            ncols=80,
            total=len(drifter_chunked_datasets.keys()),
            position=2,
        )

        os.makedirs(os.path.join(tmp_path, "drifters"), exist_ok=True)

        drifter_datasets = list[xr.Dataset]()
        for combine_job in as_completed(combine_jobmap.keys()):
            dataset: xr.Dataset = combine_job.result()
            drifter_datasets.append(dataset)
            bar.update()
        bar.close()
        return drifter_datasets


def to_raggedarray(
    tmp_path: str = _TMP_PATH,
    skip_download: bool = False,
    max: int | None = None,
    chunk_size: int = 100_000,
    use_fill_values: bool = True,
    max_chunks: int | None = None,
) -> xr.Dataset:
    """Get the GDP source dataset."""

    os.makedirs(tmp_path, exist_ok=True)

    requests = _get_download_list(tmp_path)

    # Filter down for testing purposes.
    if max:
        requests = [requests[max]]

    # Download necessary data and metadata files.
    if not skip_download:
        download_with_progress(requests)

    gdp_metadata_df = get_gdp_metadata(tmp_path)

    # Run async process to parallelize data processing.
    drifter_datasets = asyncio.run(
        _parallel_get(
            [dst for (_, dst) in requests],
            gdp_metadata_df,
            chunk_size,
            tmp_path,
            use_fill_values,
            max_chunks,
        )
    )

    # Sort the drifters by their start date.
    deploy_date_id_map = {
        ds["id"].data[0]: ds["start_date"].data[0] for ds in drifter_datasets
    }
    deploy_date_sort_key = np.argsort(list(deploy_date_id_map.values()))
    sorted_drifter_datasets = [drifter_datasets[idx] for idx in deploy_date_sort_key]

    # Concatenate drifter data and metadata variables separately.
    obs_ds = xr.concat(
        [ds.drop_dims("traj") for ds in sorted_drifter_datasets],
        dim="obs",
        data_vars=_DATA_VARS,
    )
    traj_ds = xr.concat(
        [ds.drop_dims("obs") for ds in sorted_drifter_datasets],
        dim="traj",
        data_vars=_METADATA_VARS,
    )

    # Merge the separate datasets.
    agg_ds = xr.merge([obs_ds, traj_ds])

    # Add variable metadata.
    for var_name in _DATA_VARS + _METADATA_VARS:
        if var_name in VARS_ATTRS.keys():
            agg_ds[var_name].attrs = VARS_ATTRS[var_name]
    agg_ds.attrs = ATTRS

    return agg_ds



import unittest
from concurrent import futures
from datetime import datetime, timedelta
from typing import Any

import numpy as np
import pandas as pd
import xarray as xr

from clouddrift.kinematics import velocity_from_position
from clouddrift.ragged import (
    apply_ragged,
    chunk,
    obs_index_to_row,
    prune,
    ragged_to_regular,
    regular_to_ragged,
    segment,
    subset,
    unpack,
)
from clouddrift.raggedarray import RaggedArray

if __name__ == "__main__":
    unittest.main()


def sample_ragged_array() -> RaggedArray:
    drifter_id = [1, 3, 2]
    longitude = [[-121, -111, 51, 61, 71], [103, 113], [12, 22, 32, 42]]
    latitude = [[-90, -45, 45, 90, 0], [10, 20], [10, 20, 30, 40]]
    t = [[1, 2, 3, 4, 5], [4, 5], [2, 3, 4, 5]]
    test = [
        [True, True, True, False, False],
        [False, False],
        [True, False, False, False],
    ]
    rowsize = [len(x) for x in longitude]
    attrs_global = {
        "title": "test trajectories",
        "history": "version xyz",
    }
    coords: dict[str, list] = {"id": drifter_id, "time": t}
    metadata = {"rowsize": rowsize}
    data: dict[str, list] = {"test": test, "lat": latitude, "lon": longitude}

    # append xr.Dataset to a list
    list_ds = []
    for i in range(0, len(rowsize)):
        xr_coords = {}
        xr_coords["id"] = (
            ["rows"],
            [coords["id"][i]],
            {"long_name": "variable id", "units": "-"},
        )
        xr_coords["time"] = (
            ["obs"],
            coords["time"][i],
            {"long_name": "variable time", "units": "-"},
        )

        xr_data: dict[str, Any] = {}
        for var in metadata.keys():
            xr_data[var] = (
                ["traj"],
                [metadata[var][i]],
                {"long_name": f"variable {var}", "units": "-"},
            )

        for var in data.keys():
            xr_data[var] = (
                ["obs"],
                data[var][i],
                {"long_name": f"variable {var}", "units": "-"},
            )

        list_ds.append(
            xr.Dataset(coords=xr_coords, data_vars=xr_data, attrs=attrs_global)
        )

    ra = RaggedArray.from_files(
        [0, 1, 2],
        lambda i: list_ds[i],
        ["id", "time"],
        name_meta=["rowsize"],
        name_data=["test", "lat", "lon"],
        name_dims={"rows": "rows", "obs": "obs"},
    )

    return ra


class chunk_tests(unittest.TestCase):
    def test_chunk(self):
        # Simple chunk without trimming
        self.assertTrue(np.all(chunk([1, 2, 3, 4], 2) == np.array([[1, 2], [3, 4]])))

        # Simple chunk with trimming
        self.assertTrue(np.all(chunk([1, 2, 3, 4, 5], 2) == np.array([[1, 2], [3, 4]])))

        # Simple chunk with trimming skipping the first point
        self.assertTrue(
            np.all(chunk([1, 2, 3, 4, 5], 2, align="end") == np.array([[2, 3], [4, 5]]))
        )

        # Simple chunk with trimming skipping the first point
        self.assertTrue(
            np.all(
                chunk([1, 2, 3, 4, 5, 6, 7, 8], 3, align="end")
                == np.array([[3, 4, 5], [6, 7, 8]])
            )
        )

        # Simple chunk with trimming with middle alignment
        self.assertTrue(
            np.all(
                chunk([1, 2, 3, 4, 5, 6, 7, 8], 3, align="middle")
                == np.array([[2, 3, 4], [5, 6, 7]])
            )
        )

        # Simple chunk with align to the end with with overlap
        self.assertTrue(
            np.all(
                chunk([1, 2, 3, 4, 5, 6, 7, 8], 3, 1, align="end")
                == np.array([[2, 3, 4], [4, 5, 6], [6, 7, 8]])
            )
        )

        # Simple chunk with trimming skipping the first point with overlap
        self.assertTrue(
            np.all(
                chunk(np.arange(1, 12), 4, align="middle")
                == np.array([[2, 3, 4, 5], [6, 7, 8, 9]])
            )
        )

        # When length == 1, result is a transpose of the input
        self.assertTrue(np.all(chunk([1, 2, 3, 4], 1) == np.array([[1, 2, 3, 4]]).T))

        # When length > len(x), result is an empty 2-d array
        self.assertTrue(chunk([1], 2).shape == (0, 2))

        # When length == 0, the function raises a ZeroDivisionError
        with self.assertRaises(ZeroDivisionError):
            chunk([1], 0)

        # When length < 0, the function raises a ValueError
        with self.assertRaises(ValueError):
            chunk([1], -1)

        # When align is assigned a wrong value, the function raises a ValueError
        with self.assertRaises(ValueError):
            chunk([1], 1, align="wrong")

    def test_chunk_overlap(self):
        # Simple chunk with overlap
        self.assertTrue(
            np.all(
                chunk([1, 2, 3, 4, 5], 2, 1)
                == np.array([[1, 2], [2, 3], [3, 4], [4, 5]])
            )
        )

        # Overlap larger than length raises ValueError
        with self.assertRaises(ValueError):
            chunk([1, 2, 3, 4, 5], 2, 3)

        # Negative overlap offsets chunks
        self.assertTrue(
            np.all(chunk([1, 2, 3, 4, 5], 2, -1) == np.array([[1, 2], [4, 5]]))
        )

    def test_chunk_rowsize(self):
        # Simple chunk with rowsize
        self.assertTrue(
            np.all(
                apply_ragged(chunk, np.array([1, 2, 3, 4, 5, 6]), [2, 3, 1], 2)
                == np.array([[1, 2], [3, 4]])
            )
        )

        # rowsize with overlap
        self.assertTrue(
            np.all(
                apply_ragged(
                    chunk, np.array([1, 2, 3, 4, 5, 6]), [2, 3, 1], 2, overlap=1
                )
                == np.array([[1, 2], [3, 4], [4, 5]])
            )
        )

    def test_chunk_array_like(self):
        # chunk works with array-like objects
        self.assertTrue(
            np.all(chunk(np.array([1, 2, 3, 4]), 2) == np.array([[1, 2], [3, 4]]))
        )
        self.assertTrue(
            np.all(
                chunk(xr.DataArray(data=[1, 2, 3, 4]), 2) == np.array([[1, 2], [3, 4]])
            )
        )
        self.assertTrue(
            np.all(chunk(pd.Series(data=[1, 2, 3, 4]), 2) == np.array([[1, 2], [3, 4]]))
        )


class prune_tests(unittest.TestCase):
    def test_prune(self):
        x = [1, 2, 3, 1, 2, 1, 2, 3, 4]
        rowsize = [3, 2, 4]
        minimum = 3

        for data in [x, np.array(x), pd.Series(data=x), xr.DataArray(data=x)]:
            x_new, rowsize_new = prune(data, rowsize, minimum)
            self.assertTrue(isinstance(x_new, np.ndarray))
            self.assertTrue(isinstance(rowsize_new, np.ndarray))
            np.testing.assert_equal(x_new, [1, 2, 3, 1, 2, 3, 4])
            np.testing.assert_equal(rowsize_new, [3, 4])

    def test_prune_all_longer(self):
        x = [1, 2, 3, 1, 2, 1, 2, 3, 4]
        rowsize = [3, 2, 4]
        minimum = 1

        for data in [x, np.array(x), pd.Series(data=x), xr.DataArray(data=x)]:
            x_new, rowsize_new = prune(data, rowsize, minimum)
            np.testing.assert_equal(x_new, data)
            np.testing.assert_equal(rowsize_new, rowsize)

    def test_prune_all_smaller(self):
        x = [1, 2, 3, 1, 2, 1, 2, 3, 4]
        rowsize = [3, 2, 4]
        minimum = 5

        for data in [x, np.array(x), pd.Series(data=x), xr.DataArray(data=x)]:
            x_new, rowsize_new = prune(data, rowsize, minimum)
            np.testing.assert_equal(x_new, np.array([]))
            np.testing.assert_equal(rowsize_new, np.array([]))

    def test_prune_dates(self):
        a = pd.date_range(
            start=pd.to_datetime("1/1/2018"),
            end=pd.to_datetime("1/03/2018"),
        )

        b = pd.date_range(
            start=pd.to_datetime("1/1/2018"),
            end=pd.to_datetime("1/05/2018"),
        )

        c = pd.date_range(
            start=pd.to_datetime("1/1/2018"),
            end=pd.to_datetime("1/08/2018"),
        )

        x = np.concatenate((a, b, c))
        rowsize = [len(v) for v in [a, b, c]]

        x_new, rowsize_new = prune(x, rowsize, 5)
        np.testing.assert_equal(x_new, np.concatenate((b, c)))
        np.testing.assert_equal(rowsize_new, [5, 8])

    def test_prune_keep_nan(self):
        x = [1, 2, np.nan, 1, 2, 1, 2, np.nan, 4]
        rowsize = [3, 2, 4]
        minimum = 3

        for data in [x, np.array(x), pd.Series(data=x), xr.DataArray(data=x)]:
            x_new, rowsize_new = prune(data, rowsize, minimum)
            np.testing.assert_equal(x_new, [1, 2, np.nan, 1, 2, np.nan, 4])
            np.testing.assert_equal(rowsize_new, [3, 4])

    def test_prune_empty(self):
        x = []
        rowsize = []
        minimum = 3

        for data in [x, np.array(x), pd.Series(data=x), xr.DataArray(data=x)]:
            with self.assertRaises(IndexError):
                x_new, rowsize_new = prune(data, rowsize, minimum)

    def test_print_incompatible_rowsize(self):
        x = [1, 2, 3, 1, 2]
        rowsize = [3, 3]
        minimum = 3

        for data in [x, np.array(x), pd.Series(data=x), xr.DataArray(data=x)]:
            with self.assertRaises(ValueError):
                x_new, rowsize_new = prune(data, rowsize, minimum)


class segment_tests(unittest.TestCase):
    def test_segment(self):
        x = [0, 1, 1, 1, 2, 2, 3, 3, 3, 3, 4]
        tol = 0.5
        self.assertTrue(isinstance(segment(x, tol), np.ndarray))
        self.assertTrue(np.all(segment(x, tol) == np.array([1, 3, 2, 4, 1])))
        self.assertTrue(np.all(segment(np.array(x), tol) == np.array([1, 3, 2, 4, 1])))
        self.assertTrue(
            np.all(segment(xr.DataArray(data=x), tol) == np.array([1, 3, 2, 4, 1]))
        )

    def test_segment_zero_tolerance(self):
        x = [1, 2, 2, 3, 3, 3, 4, 4, 4, 4]
        tol = 0
        self.assertIsNone(
            np.testing.assert_equal(segment(x, tol), np.array([1, 2, 3, 4]))
        )

    def test_segment_negative_tolerance(self):
        x = [0, 1, 1, 1, 2, 0, 3, 3, 3, 4]
        tol = -1
        self.assertTrue(np.all(segment(x, tol) == np.array([5, 5])))

    def test_segment_rowsize(self):
        x = [0, 1, 1, 1, 2, 2, 3, 3, 3, 3, 4]
        tol = 0.5
        rowsize = [6, 5]
        segment_sizes = segment(x, tol, rowsize)
        self.assertTrue(isinstance(segment_sizes, np.ndarray))
        self.assertTrue(np.all(segment_sizes == np.array([1, 3, 2, 4, 1])))

    def test_segment_positive_and_negative_tolerance(self):
        x = [1, 1, 2, 2, 1, 1, 2, 2]
        segment_sizes = segment(x, 0.5, rowsize=segment(x, -0.5))
        self.assertTrue(np.all(segment_sizes == np.array([2, 2, 2, 2])))

    def test_segment_rowsize_raises(self):
        x = [0, 1, 2, 3]
        tol = 0.5
        rowsize = [1, 2]  # rowsize is too short
        with self.assertRaises(ValueError):
            segment(x, tol, rowsize)

    def test_segments_datetime(self):
        x = [
            datetime(2023, 1, 1),
            datetime(2023, 1, 2),
            datetime(2023, 1, 3),
            datetime(2023, 2, 1),
            datetime(2023, 2, 2),
        ]
        for tol in [pd.Timedelta("1 day"), timedelta(days=1), np.timedelta64(1, "D")]:
            self.assertIsNone(
                np.testing.assert_equal(segment(x, tol), np.array([3, 2]))
            )

    def test_segments_numpy(self):
        x = np.array(
            [
                np.datetime64("2023-01-01"),
                np.datetime64("2023-01-02"),
                np.datetime64("2023-01-03"),
                np.datetime64("2023-02-01"),
                np.datetime64("2023-02-02"),
            ]
        )
        for tol in [pd.Timedelta("1 day"), timedelta(days=1), np.timedelta64(1, "D")]:
            self.assertIsNone(
                np.testing.assert_equal(segment(x, tol), np.array([3, 2]))
            )

    def test_segments_pandas(self):
        x = pd.to_datetime(["1/1/2023", "1/2/2023", "1/3/2023", "2/1/2023", "2/2/2023"])
        for tol in [pd.Timedelta("1 day"), timedelta(days=1), np.timedelta64(1, "D")]:
            self.assertIsNone(
                np.testing.assert_equal(segment(x, tol), np.array([3, 2]))
            )


class ragged_to_regular_tests(unittest.TestCase):
    def test_ragged_to_regular(self):
        ragged = np.array([1, 2, 3, 4, 5])
        rowsize = [2, 1, 2]
        expected = np.array([[1, 2], [3, np.nan], [4, 5]])

        result = ragged_to_regular(ragged, rowsize)
        self.assertTrue(np.all(np.isnan(result) == np.isnan(expected)))
        self.assertTrue(
            np.all(result[~np.isnan(result)] == expected[~np.isnan(expected)])
        )

        result = ragged_to_regular(
            xr.DataArray(data=ragged), xr.DataArray(data=rowsize)
        )
        self.assertTrue(np.all(np.isnan(result) == np.isnan(expected)))
        self.assertTrue(
            np.all(result[~np.isnan(result)] == expected[~np.isnan(expected)])
        )

        result = ragged_to_regular(pd.Series(data=ragged), pd.Series(data=rowsize))
        self.assertTrue(np.all(np.isnan(result) == np.isnan(expected)))
        self.assertTrue(
            np.all(result[~np.isnan(result)] == expected[~np.isnan(expected)])
        )

    def test_ragged_to_regular_fill_value(self):
        ragged = np.array([1, 2, 3, 4, 5])
        rowsize = [2, 1, 2]
        expected = np.array([[1, 2], [3, -999], [4, 5]])

        result = ragged_to_regular(ragged, rowsize, fill_value=-999)
        self.assertTrue(np.all(result == expected))

    def test_regular_to_ragged(self):
        regular = np.array([[1, 2], [3, np.nan], [4, 5]])
        expected = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        expected_rowsize = np.array([2, 1, 2])

        result, rowsize = regular_to_ragged(regular)
        self.assertTrue(np.all(result == expected))
        self.assertTrue(np.all(rowsize == expected_rowsize))

    def test_regular_to_ragged_fill_value(self):
        regular = np.array([[1, 2], [3, -999], [4, 5]])
        expected = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        expected_rowsize = np.array([2, 1, 2])

        result, rowsize = regular_to_ragged(regular, fill_value=-999)
        self.assertTrue(np.all(result == expected))
        self.assertTrue(np.all(rowsize == expected_rowsize))

    def test_ragged_to_regular_roundtrip(self):
        ragged = np.array([1, 2, 3, 4, 5])
        rowsize = [2, 1, 2]
        new_ragged, new_rowsize = regular_to_ragged(ragged_to_regular(ragged, rowsize))
        self.assertTrue(np.all(new_ragged == ragged))
        self.assertTrue(np.all(new_rowsize == rowsize))


class apply_ragged_tests(unittest.TestCase):
    def setUp(self):
        self.rowsize = [2, 3, 4]
        self.x = np.array([1, 2, 10, 12, 14, 30, 33, 36, 39])
        self.y = np.arange(0, len(self.x))
        self.t = np.array([1, 2, 1, 2, 3, 1, 2, 3, 4])

    def test_simple(self):
        y = apply_ragged(
            lambda x: x**2,
            np.array([1, 2, 3, 4]),
            [2, 2],
        )
        self.assertTrue(np.all(y == np.array([1, 4, 9, 16])))

    def test_simple_dataarray(self):
        y = apply_ragged(
            lambda x: x**2,
            xr.DataArray(data=[1, 2, 3, 4], coords={"obs": [1, 2, 3, 4]}),
            [2, 2],
        )
        self.assertTrue(np.all(y == np.array([1, 4, 9, 16])))

    def test_simple_with_args(self):
        y = apply_ragged(
            lambda x, p: x**p,
            np.array([1, 2, 3, 4]),
            [2, 2],
            2,
        )
        self.assertTrue(np.all(y == np.array([1, 4, 9, 16])))

    def test_simple_with_kwargs(self):
        y = apply_ragged(
            lambda x, p: x**p,
            np.array([1, 2, 3, 4]),
            [2, 2],
            p=2,
        )
        self.assertTrue(np.all(y == np.array([1, 4, 9, 16])))

    def test_with_rows(self):
        y = apply_ragged(
            lambda x: x**2,
            np.array([1, 2, 3, 4]),
            [2, 2],
            rows=0,
        )
        self.assertTrue(np.all(y == np.array([1, 4])))

        y = apply_ragged(
            lambda x: x**2,
            np.array([1, 2, 3, 4]),
            [2, 2],
            rows=[0, 1],
        )
        self.assertTrue(np.all(y == np.array([1, 4, 9, 16])))

    def test_with_axis(self):
        # Test that axis=0 is the default.
        x = np.arange(6).reshape((3, 2))
        func = lambda x: x**2
        rowsize = [2, 1]
        y = apply_ragged(func, x, rowsize, axis=0)
        self.assertTrue(np.all(y == apply_ragged(func, x, rowsize)))

        # Test that the rowsize is checked against the correct axis.
        with self.assertRaises(ValueError):
            y = apply_ragged(func, x.T, rowsize, axis=0)

        # Test that applying an element-wise function on a 2-d array over
        # ragged axis 1 is th same as applying it to the transpose over ragged
        # axis 0.
        rowsize = [1, 1]
        y0 = apply_ragged(func, x.T, rowsize, axis=0)
        y1 = apply_ragged(func, x, rowsize, axis=1)
        self.assertTrue(np.all(y0 == y1.T))

        # Test that axis=1 works with reduction over the non-ragged axis.
        y = apply_ragged(np.sum, x, rowsize, axis=1)
        self.assertTrue(np.all(y == np.sum(x, axis=0)))

        # Test that the same works with xr.DataArray as input
        # (this did not work before the axis feature was added).
        y = apply_ragged(np.sum, xr.DataArray(data=x), rowsize, axis=1)
        self.assertTrue(np.all(y == np.sum(x, axis=0)))

        # Test that axis works for multiple outputs:
        func = lambda x: (np.mean(x), np.std(x))
        y = apply_ragged(func, x, rowsize, axis=1)
        self.assertTrue(np.all(y[0] == np.mean(x, axis=0)))
        self.assertTrue(np.all(y[1] == np.std(x, axis=0)))

    def test_velocity_ndarray(self):
        for executor in [futures.ThreadPoolExecutor(), futures.ProcessPoolExecutor()]:
            u, v = apply_ragged(
                velocity_from_position,
                [self.x, self.y, self.t],
                self.rowsize,
                coord_system="cartesian",
                executor=executor,
            )
            self.assertIsNone(
                np.testing.assert_allclose(
                    u, [1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0]
                )
            )
            self.assertIsNone(
                np.testing.assert_allclose(
                    v, [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
                )
            )

    def test_velocity_dataarray(self):
        for executor in [futures.ThreadPoolExecutor(), futures.ProcessPoolExecutor()]:
            u, v = apply_ragged(
                velocity_from_position,
                [
                    xr.DataArray(data=self.x),
                    xr.DataArray(data=self.y),
                    xr.DataArray(data=self.t),
                ],
                xr.DataArray(data=self.rowsize),
                coord_system="cartesian",
                executor=executor,
            )
            self.assertIsNone(
                np.testing.assert_allclose(
                    u, [1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0]
                )
            )
            self.assertIsNone(
                np.testing.assert_allclose(
                    v, [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
                )
            )

    def test_bad_rowsize_raises(self):
        with self.assertRaises(ValueError):
            for use_threads in [True, False]:
                apply_ragged(
                    lambda x: x**2,
                    np.array([1, 2, 3, 4]),
                    [2],
                    use_threads=use_threads,
                )


class subset_tests(unittest.TestCase):
    def setUp(self):
        self.ds = sample_ragged_array().to_xarray()

    def test_ds_unmodified(self):
        ds_original = self.ds.copy(deep=True)
        subset(self.ds, {"test": True})
        xr.testing.assert_equal(ds_original, self.ds)

    def test_equal(self):
        ds_sub = subset(self.ds, {"test": True})
        self.assertEqual(len(ds_sub.id), 2)

    def test_select(self):
        ds_sub = subset(self.ds, {"id": [1, 2]})
        self.assertTrue(all(ds_sub.id == [1, 2]))
        self.assertEqual(len(ds_sub.id), 2)

    def test_range(self):
        # positive
        ds_sub = subset(self.ds, {"lon": (0, 180)})
        traj_idx = np.insert(np.cumsum(ds_sub["rowsize"].values), 0, 0)
        self.assertTrue(
            all(ds_sub.lon[slice(traj_idx[0], traj_idx[1])] == [51, 61, 71])
        )

        self.assertTrue(all(ds_sub.lon[slice(traj_idx[1], traj_idx[2])] == [103, 113]))

        self.assertTrue(
            all(ds_sub.lon[slice(traj_idx[2], traj_idx[3])] == [12, 22, 32, 42])
        )

        # negative range
        ds_sub = subset(self.ds, {"lon": (-180, 0)})
        traj_idx = np.insert(np.cumsum(ds_sub["rowsize"].values), 0, 0)
        self.assertEqual(len(ds_sub.id), 1)
        self.assertEqual(ds_sub.id[0], 1)
        self.assertTrue(all(ds_sub.lon == [-121, -111]))

        # both
        ds_sub = subset(self.ds, {"lon": (-30, 30)})
        traj_idx = np.insert(np.cumsum(ds_sub["rowsize"].values), 0, 0)
        self.assertEqual(len(ds_sub.id), 1)
        self.assertEqual(ds_sub.id[0], 2)
        self.assertTrue(all(ds_sub.lon[slice(traj_idx[0], traj_idx[1])] == ([12, 22])))

    def test_combine(self):
        ds_sub = subset(
            self.ds, {"id": [1, 2], "lat": (-90, 20), "lon": (-180, 25), "test": True}
        )
        self.assertTrue(all(ds_sub.id == [1, 2]))
        self.assertTrue(all(ds_sub.lon == [-121, -111, 12]))
        self.assertTrue(all(ds_sub.lat == [-90, -45, 10]))

    def test_empty(self):
        with self.assertWarns(UserWarning):
            ds_sub = subset(self.ds, {"id": 3, "lon": (-180, 0)})
            self.assertTrue(len(ds_sub.sizes) == 0)

    def test_unknown_var(self):
        with self.assertRaises(ValueError):
            subset(self.ds, {"a": 10})

        with self.assertRaises(ValueError):
            subset(self.ds, {"lon": (0, 180), "a": (0, 10)})

    def test_ragged_array_with_id_as_str(self):
        ds_str = self.ds.copy()
        ds_str["id"].values = ds_str["id"].astype(str)

        ds_sub = subset(ds_str, {"id": ds_str["id"].values[0]})
        self.assertTrue(ds_sub["id"].size == 1)

        ds_sub = subset(ds_str, {"id": list(ds_str["id"].values[:2])})
        self.assertTrue(ds_sub["id"].size == 2)

    def test_ragged_array_with_id_as_object(self):
        ds_str = self.ds.copy()
        ds_str["id"].values = ds_str["id"].astype(object)

        ds_sub = subset(ds_str, {"id": ds_str["id"].values[0]})
        self.assertTrue(ds_sub["id"].size == 1)

        ds_sub = subset(ds_str, {"id": list(ds_str["id"].values[:2])})
        self.assertTrue(ds_sub["id"].size == 2)

    def test_arraylike_criterion(self):
        # DataArray
        ds_sub = subset(self.ds, {"id": self.ds["id"][:2]})
        self.assertTrue(ds_sub["id"].size == 2)

        # NumPy array
        ds_sub = subset(self.ds, {"id": self.ds["id"][:2].values})
        self.assertTrue(ds_sub["id"].size == 2)

    def test_full_rows(self):
        ds_id_rowsize = {
            i: j for i, j in zip(self.ds.id.values, self.ds.rowsize.values)
        }

        ds_sub = subset(self.ds, {"lon": (-125, -111)}, full_rows=True)
        self.assertTrue(all(ds_sub.lon == [-121, -111, 51, 61, 71]))

        ds_sub_id_rowsize = {
            i: j for i, j in zip(ds_sub.id.values, ds_sub.rowsize.values)
        }
        for k, v in ds_sub_id_rowsize.items():
            self.assertTrue(ds_id_rowsize[k] == v)

        ds_sub = subset(self.ds, {"lat": (30, 40)}, full_rows=True)
        self.assertTrue(all(ds_sub.lat == [10, 20, 30, 40]))

        ds_sub_id_rowsize = {
            i: j for i, j in zip(ds_sub.id.values, ds_sub.rowsize.values)
        }
        for k, v in ds_sub_id_rowsize.items():
            self.assertTrue(ds_id_rowsize[k] == v)

        ds_sub = subset(self.ds, {"time": (4, 5)}, full_rows=True)
        xr.testing.assert_equal(self.ds, ds_sub)

    def test_subset_by_rows(self):
        rows = [0, 2]  # test extracting first and third rows
        ds_sub = subset(self.ds, {"rows": rows})
        self.assertTrue(all(ds_sub["id"] == [1, 2]))
        self.assertTrue(all(ds_sub["rowsize"] == [5, 4]))

    def test_subset_callable(self):
        func = (
            lambda arr: ((arr - arr[0]) % 2) == 0
        )  # test keeping obs every two time intervals
        ds_sub = subset(self.ds, {"time": func})
        self.assertTrue(all(ds_sub["id"] == [1, 3, 2]))
        self.assertTrue(all(ds_sub["rowsize"] == [3, 1, 2]))
        self.assertTrue(all(ds_sub["time"] == [1, 3, 5, 4, 2, 4]))

        func = lambda arr: arr <= 2  # keep id larger or equal to 2
        ds_sub = subset(self.ds, {"id": func})
        self.assertTrue(all(ds_sub["id"] == [1, 2]))
        self.assertTrue(all(ds_sub["rowsize"] == [5, 4]))

    def test_subset_callable_tuple(self):
        func = lambda arr1, arr2: np.logical_and(
            arr1 >= 0, arr2 >= 30
        )  # keep positive longitude and latitude larger or equal than 30
        ds_sub = subset(self.ds, {("lon", "lat"): func})
        self.assertTrue(all(ds_sub["id"] == [1, 2]))
        self.assertTrue(all(ds_sub["rowsize"] == [2, 2]))
        self.assertTrue(all(ds_sub["lon"] >= 0))
        self.assertTrue(all(ds_sub["lat"] >= 30))

    def test_subset_callable_wrong_dim(self):
        func = lambda arr: [arr, arr]  # returns 2 values per element
        with self.assertRaises(ValueError):
            subset(self.ds, {"time": func})
        with self.assertRaises(ValueError):
            subset(self.ds, {"id": func})

    def test_subset_callable_wrong_type(self):
        rows = [0, 2]  # test extracting first and third rows
        with self.assertRaises(TypeError):  # passing a tuple when a string is expected
            subset(self.ds, {("rows",): rows})

    def test_subset_callable_tuple_unknown_var(self):
        func = lambda arr1, arr2: np.logical_and(
            arr1 >= 0, arr2 >= 30
        )  # keep positive longitude and latitude larger or equal than 30
        with self.assertRaises(ValueError):
            subset(self.ds, {("a", "lat"): func})

    def test_subset_callable_tuple_not_same_dimension(self):
        func = lambda arr1, arr2: np.logical_and(
            arr1 >= 0, arr2 >= 30
        )  # keep positive longitude and latitude larger or equal than 30
        with self.assertRaises(TypeError):
            subset(self.ds, {("id", "lat"): func})


class unpack_tests(unittest.TestCase):
    def test_unpack(self):
        ds = sample_ragged_array().to_xarray()

        # Test unpacking into DataArrays
        lon = unpack(ds.lon, ds["rowsize"])

        self.assertTrue(isinstance(lon, list))
        self.assertTrue(np.all([isinstance(a, xr.DataArray) for a in lon]))
        self.assertTrue(
            np.all([lon[n].size == ds["rowsize"][n] for n in range(len(lon))])
        )

        # Test unpacking into np.ndarrays
        lon = unpack(ds.lon.values, ds["rowsize"])

        self.assertTrue(isinstance(lon, list))
        self.assertTrue(np.all([isinstance(a, np.ndarray) for a in lon]))
        self.assertTrue(
            np.all([lon[n].size == ds["rowsize"][n] for n in range(len(lon))])
        )

    def test_unpack_rows(self):
        ds = sample_ragged_array().to_xarray()
        x = ds.lon.values
        rowsize = ds.rowsize.values

        self.assertTrue(
            all(
                np.array_equal(a, b)
                for a, b in zip(unpack(x, rowsize, None), unpack(x, rowsize))
            )
        )
        self.assertTrue(
            all(
                np.array_equal(a, b)
                for a, b in zip(unpack(x, rowsize, 0), unpack(x, rowsize)[:1])
            )
        )
        self.assertTrue(
            all(
                np.array_equal(a, b)
                for a, b in zip(unpack(x, rowsize, [0, 1]), unpack(x, rowsize)[:2])
            )
        )
        # Test that unpack can accept rows as numpy integer as well, not just
        # the built-in int.
        self.assertTrue(
            all(
                np.array_equal(a, b)
                for a, b in zip(unpack(x, rowsize, np.int64(0)), unpack(x, rowsize)[:1])
            )
        )


class obs_index_to_row_tests(unittest.TestCase):
    def test_obs_index_to_row(self):
        index = list(range(10))
        rowsize = [2, 5, 3]
        row = obs_index_to_row(index, rowsize)
        self.assertTrue(np.all(row == [0, 0, 1, 1, 1, 1, 1, 2, 2, 2]))

    def test_obs_index_to_row_array_like_rowsize(self):
        index = list(range(10))
        rowsize = xr.DataArray(data=[2, 5, 3])
        row = obs_index_to_row(index, rowsize)
        self.assertTrue(np.all(row == np.array([0, 0, 1, 1, 1, 1, 1, 2, 2, 2])))

    def test_obs_index_to_row_array_rowsize(self):
        index = list(range(10))
        rowsize = np.array([2, 5, 3])
        row = obs_index_to_row(index, rowsize)
        self.assertTrue(np.all(row == np.array([0, 0, 1, 1, 1, 1, 1, 2, 2, 2])))

    def test_obs_index_to_row_array_like_index(self):
        index = xr.DataArray(data=list(range(10)))
        rowsize = [2, 5, 3]
        row = obs_index_to_row(index, rowsize)
        self.assertTrue(np.all(row == np.array([0, 0, 1, 1, 1, 1, 1, 2, 2, 2])))

    def test_obs_index_to_row_array_index(self):
        index = np.array(range(10))
        rowsize = [2, 5, 3]
        row = obs_index_to_row(index, rowsize)
        self.assertTrue(np.all(row == np.array([0, 0, 1, 1, 1, 1, 1, 2, 2, 2])))

    def test_obs_index_to_row_array_like(self):
        index = xr.DataArray(data=list(range(10)))
        rowsize = xr.DataArray(data=[2, 5, 3])
        row = obs_index_to_row(index, rowsize)
        self.assertTrue(np.all(row == np.array([0, 0, 1, 1, 1, 1, 1, 2, 2, 2])))

    def test_obs_index_to_row_array(self):
        index = np.array(range(10))
        rowsize = np.array([2, 5, 3])
        row = obs_index_to_row(index, rowsize)
        self.assertTrue(np.all(row == np.array([0, 0, 1, 1, 1, 1, 1, 2, 2, 2])))

    def test_obs_index_to_row_out_of_bounds(self):
        index = 10
        rowsize = [2, 5, 3]
        with self.assertRaises(ValueError):
            obs_index_to_row(index, rowsize)

    def test_obs_index_to_row_negative(self):
        index = -1
        rowsize = [2, 5, 3]
        with self.assertRaises(ValueError):
            obs_index_to_row(index, rowsize)



from collections.abc import Sequence
from unittest.mock import Mock, _patch


class MultiPatcher:
    _patches: Sequence[_patch]

    def __init__(self, patches: Sequence[_patch]):
        self._patches = patches

    def __enter__(self) -> Sequence[Mock]:
        return [p.start() for p in self._patches]

    def __exit__(self, *_):
        [p.stop() for p in self._patches]



"""
This module defines functions used to adapt the YoMaHa'07: Velocity data assessed
from trajectories of Argo floats at parking level and at the sea surface as
a ragged-arrays dataset.

The dataset is hosted at http://apdrc.soest.hawaii.edu/projects/yomaha/ and the user manual
is available at http://apdrc.soest.hawaii.edu/projects/yomaha/yomaha07/YoMaHa070612.pdf.

Example
-------
>>> from clouddrift.adapters import yomaha
>>> ds = yomaha.to_xarray()

Reference
---------
Lebedev, K. V., Yoshinari, H., Maximenko, N. A., & Hacker, P. W. (2007). Velocity data assessed from trajectories of Argo floats at parking level and at the sea surface. IPRC Technical Note, 4(2), 1-16.
"""

import gzip
import logging
import os
import shutil
import tempfile
import warnings
from datetime import datetime
from io import BytesIO

import numpy as np
import pandas as pd
import xarray as xr

from clouddrift.adapters.utils import download_with_progress

_logger = logging.getLogger(__name__)
YOMAHA_URLS = [
    # order of the URLs is important
    "http://apdrc.soest.hawaii.edu/projects/Argo/data/trjctry/float_types.txt",
    "http://apdrc.soest.hawaii.edu/projects/Argo/data/trjctry/DACs.txt",
    "http://apdrc.soest.hawaii.edu/projects/Argo/data/trjctry/0-Near-Real_Time/0-date_time.txt",
    "http://apdrc.soest.hawaii.edu/projects/Argo/data/trjctry/0-Near-Real_Time/WMO2DAC2type.txt",
    "http://apdrc.soest.hawaii.edu/projects/Argo/data/trjctry/0-Near-Real_Time/end-prog.lst",
    "http://apdrc.soest.hawaii.edu/projects/Argo/data/trjctry/0-Near-Real_Time/yomaha07.dat.gz",
]
YOMAHA_TMP_PATH = os.path.join(tempfile.gettempdir(), "clouddrift", "yomaha")


def download(tmp_path: str):
    download_requests = [
        (url, f"{tmp_path}/{url.split('/')[-1]}") for url in YOMAHA_URLS[:-1]
    ]
    download_with_progress(download_requests)

    filename_gz = f"{tmp_path}/{YOMAHA_URLS[-1].split('/')[-1]}"
    filename = filename_gz.removesuffix(".gz")

    buffer = BytesIO()
    download_with_progress([(YOMAHA_URLS[-1], buffer)])

    decompressed_fp = os.path.join(tmp_path, filename)
    with (
        open(decompressed_fp, "wb") as file,
        gzip.open(buffer, "rb") as compressed_file,
    ):
        shutil.copyfileobj(compressed_file, file)


def to_xarray(tmp_path: str | None = None):
    if tmp_path is None:
        tmp_path = YOMAHA_TMP_PATH
        os.makedirs(tmp_path, exist_ok=True)

    # get or update required files
    download(tmp_path)

    # database last update
    with open(f"{tmp_path}/{YOMAHA_URLS[2].split('/')[-1]}") as f:
        YOMAHA_VERSION = f.read().strip()
        print(f"Last database update was: {YOMAHA_VERSION}")

    # parse with panda
    col_names = [
        "lon_d",
        "lat_d",
        "pres_d",
        "time_d",
        "ve_d",
        "vn_d",
        "err_ve_d",
        "err_vn_d",
        "lon_s",
        "lat_s",
        "time_s",
        "ve_s",
        "vn_s",
        "err_ve_s",
        "err_vn_s",
        "lon_lp",
        "lat_lp",
        "time_lp",
        "lon_fc",
        "lat_fc",
        "time_fc",
        "lon_lc",
        "lat_lc",
        "time_lc",
        "surf_fix",
        "id",
        "cycle",
        "time_inv",
    ]

    na_col = list(
        map(
            lambda x: str(x),
            [
                -999.9999,
                -99.9999,
                -999.9,
                -999.999,
                -999.99,
                -999.99,
                -999.99,
                -999.99,
                -999.99,
                -99.99,
                -999.99,
                -999.99,
                -999.99,
                -999.99,
                -999.99,
                -999.99,
                -99.99,
                -999.99,
                -999.99,
                -99.99,
                -999.99,
                -999.99,
                -99.99,
                -999.99,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
            ],
        )
    )

    # open with pandas
    filename_gz = f"{tmp_path}/{YOMAHA_URLS[-1].split('/')[-1]}"
    filename = filename_gz.removesuffix(".gz")
    df = pd.read_csv(
        filename, names=col_names, sep=r"\s+", header=None, na_values=na_col
    )

    # convert to an Xarray Dataset
    ds = xr.Dataset.from_dataframe(df)

    unique_id, rowsize = np.unique(ds["id"], return_counts=True)

    # mapping of yomaha float id, wmo float id, daq id and float type
    df_wmo = pd.read_csv(
        f"{tmp_path}/{YOMAHA_URLS[3].split('/')[-1]}",
        sep=r"\s+",
        header=None,
        names=["id", "wmo_id", "dac_id", "float_type_id"],
        engine="python",
    )

    # mapping of Data Assembly Center (DAC) id and DAC name
    df_dac = pd.read_csv(
        f"{tmp_path}/{YOMAHA_URLS[1].split('/')[-1]}",
        sep=":",
        header=None,
        names=["dac_id", "dac"],
    )
    df_dac["dac"] = df_dac["dac"].str.strip()

    # mapping of float_type_id and float_type
    df_float = pd.read_csv(
        f"{tmp_path}/{YOMAHA_URLS[0].split('/')[-1]}",
        sep=":",
        header=None,
        skipfooter=4,
        names=["float_type_id", "float_type"],
        engine="python",
    )
    # there is a note on METOCEAN * in float_types.txt but the
    # float id, wmo id, and float type do not match (?)
    # so we remove the * from the type
    df_float.loc[df_float["float_type_id"] == 0, "float_type"] = "METOCEAN"

    # combine metadata
    df_metadata = (
        pd.merge(df_wmo, df_dac, on="dac_id", how="left")
        .merge(df_float, on="float_type_id", how="left")
        .loc[lambda x: np.isin(x["id"], unique_id)]
    )

    ds = (
        ds.rename_dims({"index": "obs"})
        .assign({"id": ("traj", unique_id)})
        .assign({"rowsize": ("traj", rowsize)})
        .assign({"wmo_id": ("traj", df_metadata["wmo_id"])})
        .assign({"dac_id": ("traj", df_metadata["dac_id"])})
        .assign({"float_type": ("traj", df_metadata["float_type_id"])})
        .set_coords(["id", "time_d", "time_s", "time_lp", "time_lc", "time_lp"])
        .drop_vars(["index"])
    )

    # Cast double floats to singles
    double_vars = [
        "lat_d",
        "lon_d",
        "lat_s",
        "lon_s",
        "lat_lp",
        "lon_lp",
        "lat_lc",
        "lon_lc",
    ]
    for var in [v for v in ds.variables if v not in double_vars]:
        if ds[var].dtype == "float64":
            ds[var] = ds[var].astype("float32")

    # define attributes
    vars_attrs = {
        "lon_d": {
            "long_name": "Longitude of the location where the deep velocity is calculated",
            "units": "degrees_east",
        },
        "lat_d": {
            "long_name": "Latitude of the location where the deep velocity is calculated",
            "units": "degrees_north",
        },
        "pres_d": {
            "long_name": "Reference parking pressure for this cycle",
            "units": "dbar",
        },
        "time_d": {
            "long_name": "Julian time (days) when deep velocity is estimated",
            "units": "days since 2000-01-01 00:00",
        },
        "ve_d": {
            "long_name": "Eastward component of the deep velocity",
            "units": "cm s-1",
        },
        "vn_d": {
            "long_name": "Northward component of the deep velocity",
            "units": "cm s-1",
        },
        "err_ve_d": {
            "long_name": "Error on the eastward component of the deep velocity",
            "units": "cm s-1",
        },
        "err_vn_d": {
            "long_name": "Error on the northward component of the deep velocity",
            "units": "cm s-1",
        },
        "lon_s": {
            "long_name": "Longitude of the location where the first surface velocity is calculated (over the first 6 h at surface)",
            "units": "degrees_east",
        },
        "lat_s": {
            "long_name": "Latitude of the location where the first surface velocity is calculated",
            "units": "degrees_north",
        },
        "time_s": {
            "long_name": "Julian time (days) when the first surface velocity is calculated",
            "units": "days since 2000-01-01 00:00",
        },
        "ve_s": {
            "long_name": "Eastward component of first surface velocity",
            "units": "cm s-1",
        },
        "vn_s": {
            "long_name": "Northward component of first surface velocity",
            "units": "cm s-1",
        },
        "err_ve_s": {
            "long_name": "Error on the eastward component of the first surface velocity",
            "units": "cm s-1",
        },
        "err_vn_s": {
            "long_name": "Error on the northward component of the first surface velocity",
            "units": "cm s-1",
        },
        "lon_lp": {
            "long_name": "Longitude of the last fix at the sea surface during the previous cycle",
            "units": "degrees_east",
        },
        "lat_lp": {
            "long_name": "Latitude of the last fix at the sea surface during the previous cycle",
            "units": "degrees_north",
        },
        "time_lp": {
            "long_name": "Julian time of the last fix at the sea surface during the previous cycle",
            "units": "days since 2000-01-01 00:00",
        },
        "lon_fc": {
            "long_name": "Longitude of the first fix at the sea surface during the current cycle",
            "units": "degrees_east",
        },
        "lat_fc": {
            "long_name": "Latitude of the first fix at the sea surface during the current cycle",
            "units": "degrees_north",
        },
        "time_fc": {
            "long_name": "Julian time of the first fix at the sea surface during the current cycle",
            "units": "days since 2000-01-01 00:00",
        },
        "lon_lc": {
            "long_name": "Longitude of the last fix at the sea surface during the current cycle",
            "units": "degrees_east",
        },
        "lat_lc": {
            "long_name": "Latitude of the last fix at the sea surface during the current cycle",
            "units": "degrees_north",
        },
        "time_lc": {
            "long_name": "Julian time of the last fix at the sea surface during the current cycle",
            "units": "days since 2000-01-01 00:00",
        },
        "surf_fix": {
            "long_name": "Number of surface fixes during the current cycle",
            "units": "-",
        },
        "id": {
            "long_name": "Float WMO number",
            "units": "-",
        },
        "cycle": {
            "long_name": "Cycle number",
            "units": "-",
        },
        "time_inv": {
            "long_name": "Time inversion/duplication flag",
            "description": "1 if at least one duplicate or inversion of time is found in the sequence containing last fix from the previous cycle and all fixes from the current cycle. Otherwise, 0.",
            "units": "-",
        },
        "wmo_id": {
            "long_name": "Float WMO number",
            "units": "-",
        },
        "dac_id": {
            "long_name": "Data Assembly Center (DAC) number",
            "description": "1: AOML (USA), 2: CORIOLIS (France), 3: JMA (Japan), 4: BODC (UK), 5: MEDS (Canada), 6: INCOIS (India), 7: KMA (Korea), 8: CSIRO (Australia), 9: CSIO (China)",
            "units": "-",
        },
        "float_type": {
            "long_name": "Float type",
            "description": "1: APEX, 2: SOLO, 3: PROVOR, 4: R1, 5: MARTEC, 6: PALACE, 7: NINJA, 8: NEMO, 9: ALACE, 0: METOCEAN",
            "units": "-",
        },
    }

    # global attributes
    attrs = {
        "title": "YoMaHa'07: Velocity data assessed from trajectories of Argo floats at parking level and at the sea surface",
        "history": f"Dataset updated on {YOMAHA_VERSION}",
        "date_created": datetime.now().isoformat(),
        "publisher_name": "Asia-Pacific Data Research Center",
        "publisher_url": "http://apdrc.soest.hawaii.edu/index.php",
        "license": "freely available",
    }

    # set attributes
    for var in vars_attrs.keys():
        if var in ds.keys():
            ds[var].attrs = vars_attrs[var]
        else:
            warnings.warn(f"Variable {var} not found in upstream data; skipping.")
    ds.attrs = attrs

    return ds



"""
This module defines the RaggedArray class, which is the intermediate data
structure used by CloudDrift to process custom Lagrangian datasets to Xarray
Datasets and Awkward Arrays.
"""

from __future__ import annotations

import warnings
from collections.abc import Callable
from typing import Any, Literal

import awkward as ak  # type: ignore
import numpy as np
import xarray as xr
from tqdm import tqdm

from clouddrift.ragged import rowsize_to_index

DimNames = Literal["rows", "obs"]
_DISABLE_SHOW_PROGRESS = False  # purely to de-noise our test suite output, should never be used/configured outside of that.


class RaggedArray:
    def __init__(
        self,
        coords: dict,
        metadata: dict,
        data: dict,
        attrs_global: dict = {},
        attrs_variables: dict = {},
        name_dims: dict[str, DimNames] = {},
        coord_dims: dict[str, str] = {},
    ):
        self.coords = coords
        self.coord_dims = coord_dims
        self.metadata = metadata
        self.data = data
        self.attrs_global = attrs_global
        self.attrs_variables = attrs_variables
        self.name_dims = name_dims
        self._coord_dims = coord_dims
        self.validate_attributes()

    @classmethod
    def from_awkward(
        cls,
        array: ak.Array,
        name_coords: list,
        name_dims: dict[str, DimNames],
        coord_dims: dict[str, str],
    ):
        """Load a RaggedArray instance from an Awkward Array.

        Parameters
        ----------
        array : ak.Array
            Awkward Array instance to load the data from
        name_coords : list, optional
            Names of the coordinate variables in the ragged arrays
        name_dims: dict
            Map a dimension to an alias.
        coord_dims: dict
            Map a coordinate to a dimension alias.

        Returns
        -------
        RaggedArray
            A RaggedArray instance
        """
        coords: dict[str, Any] = {}
        metadata = {}
        data = {}
        attrs_variables = {}

        attrs_global = array.layout.parameters["attrs"]

        for var in name_coords:
            alias = coord_dims[var]
            if name_dims[alias] == "obs":
                coords[var] = ak.flatten(array.obs[var]).to_numpy()
            else:
                coords[var] = array.obs[var].to_numpy()

            attrs_variables[var] = array.obs[var].layout.parameters["attrs"]

        for var in [v for v in array.fields if v != "obs"]:
            metadata[var] = array[var].to_numpy()
            attrs_variables[var] = array[var].layout.parameters["attrs"]

        for var in [v for v in array.obs.fields if v not in coords.keys()]:
            data[var] = ak.flatten(array.obs[var]).to_numpy()
            attrs_variables[var] = array.obs[var].layout.parameters["attrs"]

        return RaggedArray(
            coords, metadata, data, attrs_global, attrs_variables, name_dims, coord_dims
        )

    @classmethod
    def from_files(
        cls,
        indices: list[int],
        preprocess_func: Callable[[int], xr.Dataset],
        name_coords: list,
        name_meta: list = list(),
        name_data: list = list(),
        name_dims: dict[str, DimNames] = {},
        rowsize_func: Callable[[int], int] | None = None,
        attrs_global: dict | None = None,
        attrs_variables: dict | None = None,
        **kwargs,
    ):
        """Generate a ragged array archive from a list of files

        Parameters
        ----------
        indices : list
            Identification numbers list to iterate
        preprocess_func : Callable[[int], xr.Dataset]
            Returns a processed xarray Dataset from an identification number
        name_meta : list, optional
            Name of metadata variables to include in the archive (Defaults to [])
        name_data : list, optional
            Name of the data variables to include in the archive (Defaults to [])
        name_dims: dict
            Map an alias to a dimension.
        rowsize_func : Optional[Callable[[int], int]], optional
            Returns the number of observations from an identification number (to speed up processing) (Defaults to None)

        Returns
        -------
        RaggedArray
            A RaggedArray instance
        """
        # if no method is supplied, get the dimension from the preprocessing function
        rowsize_func = (
            rowsize_func
            if rowsize_func
            else lambda i, **kwargs: preprocess_func(i, **kwargs).sizes["obs"]
        )
        rowsize = cls.number_of_observations(rowsize_func, indices, **kwargs)
        coords, metadata, data, coord_dims = cls.allocate(
            preprocess_func,
            indices,
            rowsize,
            name_coords,
            name_meta,
            name_data,
            name_dims,
            **kwargs,
        )

        extracted_attrs_global, extracted_attrs_variables = cls.attributes(
            preprocess_func(indices[0], **kwargs),
            name_coords,
            name_meta,
            name_data,
        )

        if attrs_global is None:
            attrs_global = extracted_attrs_global

        if attrs_variables is None:
            attrs_variables = extracted_attrs_variables

        return RaggedArray(
            coords, metadata, data, attrs_global, attrs_variables, name_dims, coord_dims
        )

    @classmethod
    def from_netcdf(cls, filename: str, rows_dim_name="rows", obs_dim_name="obs"):
        """Read a ragged arrays archive from a NetCDF file.

        This is a thin wrapper around ``from_xarray()``.

        Parameters
        ----------
        filename : str
            File name of the NetCDF archive to read.

        Returns
        -------
        RaggedArray
            A ragged array instance
        """
        return cls.from_xarray(xr.open_dataset(filename), rows_dim_name, obs_dim_name)

    @classmethod
    def from_parquet(
        cls,
        filename: str,
        name_coords: list,
        name_dims: dict[str, DimNames],
        coord_dims: dict[str, str],
    ):
        """Read a ragged array from a parquet file.

        Parameters
        ----------
        filename : str
            File name of the parquet archive to read.
        name_coords : list, optional
            Names of the coordinate variables in the ragged arrays
        name_dims: dict
            Map a alias to a dimension.
        coord_dims: dict
            Map a coordinate to a dimension alias.

        Returns
        -------
        RaggedArray
            A ragged array instance
        """
        return RaggedArray.from_awkward(
            ak.from_parquet(filename), name_coords, name_dims, coord_dims
        )

    @classmethod
    def from_xarray(
        cls, ds: xr.Dataset, rows_dim_name: str = "rows", obs_dim_name: str = "obs"
    ):
        """Populate a RaggedArray instance from an xarray Dataset instance.

        Parameters
        ----------
        ds : xr.Dataset
            Xarray Dataset from which to load the RaggedArray
        rows_dim_name : str, optional
            Name of the row dimension in the xarray Dataset
        obs_dim_name : str, optional
            Name of the observations dimension in the xarray Dataset

        Returns
        -------
        RaggedArray
            A RaggedArray instance
        """
        coords = {}
        metadata = {}
        data = {}
        coord_dims = {}
        name_dims: dict[str, DimNames] = {rows_dim_name: "rows", obs_dim_name: "obs"}
        attrs_global = {}
        attrs_variables = {}

        attrs_global = ds.attrs

        for var in ds.coords.keys():
            var = str(var)
            dim = ds[var].dims[-1]
            coord_dims[var] = str(dim)
            coords[var] = ds[var].data
            attrs_variables[var] = ds[var].attrs

        for var in ds.data_vars.keys():
            if len(ds[var]) == ds.sizes.get(rows_dim_name):
                metadata[var] = ds[var].data
            elif len(ds[var]) == ds.sizes.get(obs_dim_name):
                data[var] = ds[var].data
            else:
                warnings.warn(
                    f"""
                    Variable '{var}' has unknown dimension size of
                    {len(ds[var])}, which is not rows={ds.sizes.get(rows_dim_name)} or
                    obs={ds.sizes.get(obs_dim_name)}; skipping.
                    """
                )
            attrs_variables[str(var)] = ds[var].attrs

        return RaggedArray(
            coords, metadata, data, attrs_global, attrs_variables, name_dims, coord_dims
        )

    @staticmethod
    def number_of_observations(
        rowsize_func: Callable[[int], int], indices: list, **kwargs
    ) -> np.ndarray:
        """Iterate through the files and evaluate the number of observations.

        Parameters
        ----------
        rowsize_func : Callable[[int], int]]
            Function that returns the number observations of a row from
            its identification number
        indices : list
            Identification numbers list to iterate

        Returns
        -------
        np.ndarray
            Number of observations
        """
        rowsize = np.zeros(len(indices), dtype="int")

        for i, index in tqdm(
            enumerate(indices),
            total=len(indices),
            desc="Retrieving the number of obs",
            ncols=80,
            disable=_DISABLE_SHOW_PROGRESS,
        ):
            rowsize[i] = rowsize_func(index, **kwargs)
        return rowsize

    @staticmethod
    def attributes(
        ds: xr.Dataset,
        name_coords: list,
        name_meta: list,
        name_data: list,
    ) -> tuple[dict, dict]:
        """Return global attributes and the attributes of all variables
        (name_coords, name_meta, and name_data) from an Xarray Dataset.

        Parameters
        ----------
        ds : xr.Dataset
            _description_
        name_coords : list, optional
            Name of metadata variables to include in the archive (default is [])
        name_meta : list, optional
            Name of metadata variables to include in the archive (default is [])
        name_data : list, optional
            Name of the data variables to include in the archive (default is [])

        Returns
        -------
        Tuple[dict, dict]
            The global and variables attributes
        """
        attrs_global = ds.attrs

        # coordinates, metadata, and data
        attrs_variables = {}
        for var in name_meta + name_data + name_coords:
            if var in ds.keys():
                attrs_variables[var] = ds[var].attrs
            else:
                warnings.warn(f"Variable {var} requested but not found; skipping.")

        return attrs_global, attrs_variables

    @staticmethod
    def allocate(
        preprocess_func: Callable[[int], xr.Dataset],
        indices: list,
        rowsize: list | np.ndarray | xr.DataArray,
        name_coords: list,
        name_meta: list,
        name_data: list,
        name_dims: dict[str, DimNames],
        **kwargs,
    ) -> tuple[dict, dict, dict, dict]:
        """
        Iterate through the files and fill for the ragged array associated
        with coordinates, and selected metadata and data variables.

        Parameters
        ----------
        preprocess_func : Callable[[int], xr.Dataset]
            Returns a processed xarray Dataset from an identification number.
        indices : list
            List of indices separating row in the ragged arrays.
        rowsize : list
            List of the number of observations per row.
        name_coords : list
            Name of the coordinate variables to include in the archive.
        name_meta : list, optional
            Name of metadata variables to include in the archive (Defaults to []).
        name_data : list, optional
            Name of the data variables to include in the archive (Defaults to []).
        name_dims: dict[str, DimNames]
            Dimension alias mapped to the name used by clouddrift.

        Returns
        -------
        Tuple[dict, dict, dict, dict]
            Dictionaries containing numerical data and attributes of coordinates, metadata and data variables.
        """
        # open one file to get dtype of variables
        ds = preprocess_func(indices[0], **kwargs)
        nb_rows = len(rowsize)
        nb_obs = np.sum(rowsize).astype("int")
        index_traj = rowsize_to_index(rowsize)
        dim_sizes = {}

        for alias in name_dims.keys():
            if name_dims[alias] == "rows":
                dim_sizes[alias] = nb_rows
            else:
                dim_sizes[alias] = nb_obs

        # allocate memory
        coords = {}
        coord_dims: dict[str, str] = {}
        for var in name_coords:
            dim = ds[var].dims[-1]
            dim_size = dim_sizes[dim]
            coords[var] = np.zeros(dim_size, dtype=ds[var].dtype)
            coord_dims[var] = dim

        metadata = {}
        for var in name_meta:
            try:
                metadata[var] = np.zeros(nb_rows, dtype=ds[var].dtype)
            except KeyError:
                warnings.warn(f"Variable {var} requested but not found; skipping.")

        data = {}
        for var in name_data:
            if var in ds.keys():
                data[var] = np.zeros(nb_obs, dtype=ds[var].dtype)
            else:
                warnings.warn(f"Variable {var} requested but not found; skipping.")
        ds.close()

        # loop and fill the ragged array
        for i, index in tqdm(
            enumerate(indices),
            total=len(indices),
            desc="Filling the Ragged Array",
            ncols=80,
            disable=_DISABLE_SHOW_PROGRESS,
        ):
            with preprocess_func(index, **kwargs) as ds:
                size = rowsize[i]
                oid = index_traj[i]

                for var in name_coords:
                    dim = ds[var].dims[-1]
                    if name_dims[dim] == "obs":
                        coords[var][oid : oid + size] = ds[var].data
                    else:
                        coords[var][i] = ds[var].data[0]

                for var in name_meta:
                    try:
                        metadata[var][i] = ds[var][0].data
                    except KeyError:
                        warnings.warn(
                            f"Variable {var} requested but not found; skipping."
                        )

                for var in name_data:
                    if var in ds.keys():
                        data[var][oid : oid + size] = ds[var].data
                    else:
                        warnings.warn(
                            f"Variable {var} requested but not found; skipping."
                        )

        return coords, metadata, data, coord_dims

    def validate_attributes(self):
        """Validate that each variable has an assigned attribute tag."""
        for key in (
            list(self.coords.keys())
            + list(self.metadata.keys())
            + list(self.data.keys())
        ):
            if key not in self.attrs_variables:
                self.attrs_variables[key] = {}

    def to_xarray(self):
        """Convert ragged array object to a xarray Dataset.

        Parameters
        ----------
        cast_to_float32 : bool, optional
            Cast all float64 variables to float32 (default is True). This option aims at
            minimizing the size of the xarray dataset.

        Returns
        -------
        xr.Dataset
            Xarray Dataset containing the ragged arrays and their attributes
        """
        dim_name_map = {v: k for k, v in self.name_dims.items()}

        xr_coords = {}
        for var in self.coords.keys():
            xr_coords[var] = (
                [self._coord_dims[var]],
                self.coords[var],
                self.attrs_variables[var],
            )

        xr_data = {}
        for var in self.metadata.keys():
            xr_data[var] = (
                [dim_name_map["rows"]],
                self.metadata[var],
                self.attrs_variables[var],
            )

        for var in self.data.keys():
            xr_data[var] = (
                [dim_name_map["obs"]],
                self.data[var],
                self.attrs_variables[var],
            )

        return xr.Dataset(coords=xr_coords, data_vars=xr_data, attrs=self.attrs_global)

    def to_awkward(self):
        """Convert ragged array object to an Awkward Array.

        Returns
        -------
        ak.Array
            Awkward Array containing the ragged array and its attributes
        """
        index_traj = rowsize_to_index(self.metadata["rowsize"])
        offset = ak.index.Index64(index_traj)

        data = []
        for var in self.coords.keys():
            dim = self._coord_dims[var]
            if self.name_dims[dim] == "obs":
                data.append(
                    ak.contents.ListOffsetArray(
                        offset,
                        ak.contents.NumpyArray(self.coords[var]),
                        parameters={"attrs": self.attrs_variables[var]},
                    )
                )
            else:
                data.append(
                    ak.with_parameter(
                        self.coords[var],
                        "attrs",
                        self.attrs_variables[var],
                        highlevel=False,
                    )
                )

        for var in self.data.keys():
            data.append(
                ak.contents.ListOffsetArray(
                    offset,
                    ak.contents.NumpyArray(self.data[var]),
                    parameters={"attrs": self.attrs_variables[var]},
                )
            )
        data_names = list(self.coords.keys()) + list(self.data.keys())

        metadata = []
        for var in self.metadata.keys():
            metadata.append(
                ak.with_parameter(
                    self.metadata[var],
                    "attrs",
                    self.attrs_variables[var],
                    highlevel=False,
                )
            )
        metadata_names = list(self.metadata.keys())

        # include the data inside the metadata list as a nested array
        metadata_names.append("obs")
        metadata.append(ak.Array(ak.contents.RecordArray(data, data_names)).layout)

        return ak.Array(
            ak.contents.RecordArray(
                metadata, metadata_names, parameters={"attrs": self.attrs_global}
            )
        )

    def to_netcdf(self, filename: str):
        """Export ragged array object to a NetCDF file.

        Parameters
        ----------
        filename : str
            Name of the NetCDF file to create.
        """

        self.to_xarray().to_netcdf(filename)

    def to_parquet(self, filename: str):
        """Export ragged array object to a parquet file.

        Parameters
        ----------
        filename : str
            Name of the parquet file to create.
        """
        ak.to_parquet(self.to_awkward(), filename)



import concurrent.futures
import logging
import os
import urllib
from datetime import datetime
from io import BufferedIOBase, BufferedWriter
from typing import Callable, Sequence

import requests
from requests import Response
from tenacity import (
    RetryCallState,
    WrappedFn,
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_exponential_jitter,
)
from tqdm import tqdm

_DISABLE_SHOW_PROGRESS = False  # purely to de-noise our test suite output, should never be used/configured outside of that.


def _before_call(rcs: RetryCallState):
    if rcs.attempt_number > 1:
        _logger.warn(
            f"retrying request for (fn, args): {(rcs.fn, rcs.args)}, attempt: {rcs.attempt_number}"
        )


_CHUNK_SIZE = 1_048_576  # 1MiB
_logger = logging.getLogger(__name__)

standard_retry_protocol: Callable[[WrappedFn], WrappedFn] = retry(
    retry=retry_if_exception(
        lambda ex: isinstance(
            ex,
            (
                requests.Timeout,
                requests.ConnectionError,
                urllib.error.HTTPError,
                urllib.error.URLError,
            ),
        )
    ),
    wait=wait_exponential_jitter(
        initial=1.25, jitter=10
    ),  # ~ 20-25 minutes total time before completely failing
    stop=stop_after_attempt(10),
    before=_before_call,
)


def download_with_progress(
    download_map: Sequence[
        tuple[str, BufferedIOBase | str] | tuple[str, BufferedIOBase | str, float]
    ],
    show_list_progress: bool | None = None,
    desc: str = "Downloading files",
    custom_retry_protocol: Callable[[WrappedFn], WrappedFn] | None = None,
):
    if show_list_progress is None:
        show_list_progress = len(download_map) > 20
    if custom_retry_protocol is None:
        retry_protocol = standard_retry_protocol
    else:
        retry_protocol = custom_retry_protocol  # type: ignore

    buffer: BufferedIOBase | BufferedWriter
    executor = concurrent.futures.ThreadPoolExecutor()
    futures: dict[
        concurrent.futures.Future,
        tuple[str, BufferedIOBase | str, BufferedIOBase | BufferedWriter],
    ] = dict()
    bar = None

    for request in download_map:
        if len(request) > 2:
            src, dst, exp_size = request
        else:
            src, dst = request
            exp_size = None

        if isinstance(dst, (str,)):
            buffer = open(dst, "wb")
        else:
            buffer = dst

        futures[
            executor.submit(
                retry_protocol(_download_with_progress),
                src,
                buffer,
                exp_size,
                not show_list_progress,
            )
        ] = (src, dst, buffer)

    try:
        if show_list_progress:
            bar = tqdm(
                desc=desc,
                total=len(futures),
                unit="Files",
                disable=_DISABLE_SHOW_PROGRESS,
            )

        for fut in concurrent.futures.as_completed(futures):
            src, dst, buffer = futures[fut]

            if isinstance(dst, (str,)):
                buffer.close()

            ex = fut.exception(0)
            if ex is None:
                _logger.debug(f"Finished download job: ({src}, {dst})")
                if bar is not None:
                    bar.update(1)
            else:
                raise ex
    except Exception as e:
        _logger.error(
            f"Got the following exception: {str(e)}, cancelling all other jobs and cleaning up \
              any created resources."
        )
        for x in futures.keys():
            src, dst, buffer = futures[x]

            if not x.done():
                x.cancel()

            if isinstance(dst, (str,)) and os.path.exists(dst):
                os.remove(dst)
        raise e
    finally:
        executor.shutdown(True)
        if bar is not None:
            bar.close()


def _download_with_progress(
    url: str,
    output: BufferedIOBase | BufferedWriter,
    expected_size: float | None,
    show_progress: bool,
):
    if isinstance(output, str) and os.path.exists(output):
        _logger.debug(f"File exists {output} checking for updates...")
        local_last_modified = os.path.getmtime(output)

        # Get last modified time of the remote file
        with requests.head(url, timeout=5) as res:
            if "Last-Modified" in res.headers:
                remote_last_modified = datetime.strptime(
                    res.headers.get("Last-Modified", ""),
                    "%a, %d %b %Y %H:%M:%S %Z",
                )

                # compare with local modified time
                if local_last_modified >= remote_last_modified.timestamp():
                    _logger.debug(f"File: {output} is up to date; skip download.")
                    return
            else:
                _logger.warning(
                    "Cannot determine if the file has been updated on the remote source. "
                    + "'Last-Modified' header not present in server response."
                )
    _logger.debug(f"Downloading from {url} to {output}...")

    response: Response | None = None
    bar = None

    try:
        response = requests.get(url, timeout=5, stream=True)

        if (content_length := response.headers.get("Content-Length")) is not None:
            expected_size = float(content_length)

        if show_progress:
            bar = tqdm(
                desc=url,
                total=expected_size,
                unit="B",
                unit_scale=True,
                unit_divisor=_CHUNK_SIZE,
                nrows=2,
                disable=_DISABLE_SHOW_PROGRESS,
            )

        for chunk in response.iter_content(_CHUNK_SIZE):
            if not chunk:
                break
            output.write(chunk)
            if bar is not None:
                bar.update(len(chunk))
    finally:
        if response is not None:
            response.close()
        if bar is not None:
            bar.close()


__all__ = ["download_with_progress"]



"""
This module defines functions used to adapt the Grand LAgrangian Deployment
(GLAD) dataset as a ragged-array Xarray Dataset.

The dataset and its description are hosted at https://doi.org/10.7266/N7VD6WC8.

Example
-------
>>> from clouddrift.adapters import glad
>>> ds = glad.to_xarray()

Reference
---------
Özgökmen, Tamay. 2013. GLAD experiment CODE-style drifter trajectories (low-pass filtered, 15 minute interval records), northern Gulf of Mexico near DeSoto Canyon, July-October 2012. Distributed by: Gulf of Mexico Research Initiative Information and Data Cooperative (GRIIDC), Harte Research Institute, Texas A&M University–Corpus Christi. doi:10.7266/N7VD6WC8
"""

from io import BytesIO

import numpy as np
import pandas as pd
import xarray as xr

from clouddrift.adapters.utils import download_with_progress


def get_dataframe() -> pd.DataFrame:
    """Get the GLAD dataset as a pandas DataFrame."""
    url = "https://data.gulfresearchinitiative.org/pelagos-symfony/api/file/download/169841"
    # GRIIDC server doesn't provide Content-Length header, so we'll hardcode
    # the expected data length here.
    file_size = 155330876
    buf = BytesIO(b"")
    download_with_progress([(url, buf, file_size)])
    buf.seek(0)
    column_names = [
        "id",
        "date",
        "time",
        "latitude",
        "longitude",
        "position_error",
        "u",
        "v",
        "velocity_error",
    ]
    df = pd.read_csv(buf, sep=r"\s+", skiprows=5, names=column_names)
    df["obs"] = pd.to_datetime(df["date"] + " " + df["time"])
    df.drop(["date", "time"], axis=1, inplace=True)
    return df


def to_xarray() -> xr.Dataset:
    """Return the GLAD data as a ragged-array Xarray Dataset."""
    df = get_dataframe()
    ds = df.to_xarray()

    traj, rowsize = np.unique(ds.id, return_counts=True)

    # Make the dataset compatible with clouddrift functions.
    ds = (
        ds.swap_dims({"index": "obs"})
        .drop_vars(["id", "index"])
        .assign_coords(traj=traj)
        .assign({"rowsize": ("traj", rowsize)})
        .rename_vars({"obs": "time", "traj": "id"})
    )

    # Cast double floats to singles
    for var in ds.variables:
        if ds[var].dtype == "float64":
            ds[var] = ds[var].astype("float32")

    # Set variable attributes
    ds["longitude"].attrs = {
        "long_name": "longitude",
        "standard_name": "longitude",
        "units": "degrees_east",
    }

    ds["latitude"].attrs = {
        "long_name": "latitude",
        "standard_name": "latitude",
        "units": "degrees_north",
    }

    ds["position_error"].attrs = {
        "long_name": "position_error",
        "units": "m",
    }

    ds["u"].attrs = {
        "long_name": "eastward_sea_water_velocity",
        "standard_name": "eastward_sea_water_velocity",
        "units": "m s-1",
    }

    ds["v"].attrs = {
        "long_name": "northward_sea_water_velocity",
        "standard_name": "northward_sea_water_velocity",
        "units": "m s-1",
    }

    ds["velocity_error"].attrs = {
        "long_name": "velocity_error",
        "units": "m s-1",
    }

    # Set global attributes
    ds.attrs = {
        "title": "GLAD experiment CODE-style drifter trajectories (low-pass filtered, 15 minute interval records), northern Gulf of Mexico near DeSoto Canyon, July-October 2012",
        "institution": "Consortium for Advanced Research on Transport of Hydrocarbon in the Environment (CARTHE)",
        "source": "CODE-style drifters",
        "history": "Downloaded from https://data.gulfresearchinitiative.org/data/R1.x134.073:0004 and post-processed into a ragged-array Xarray Dataset by CloudDrift",
        "references": "Özgökmen, Tamay. 2013. GLAD experiment CODE-style drifter trajectories (low-pass filtered, 15 minute interval records), northern Gulf of Mexico near DeSoto Canyon, July-October 2012. Distributed by: Gulf of Mexico Research Initiative Information and Data Cooperative (GRIIDC), Harte Research Institute, Texas A&M University–Corpus Christi. doi:10.7266/N7VD6WC8",
    }

    return ds



import numpy as np

import tests.utils as testutils
from clouddrift import datasets
from clouddrift.ragged import apply_ragged, subset


class datasets_tests(testutils.DisableProgressTestCase):
    def test_gdp1h(self):
        with datasets.gdp1h() as ds:
            self.assertTrue(ds)

    def test_gdp6h(self):
        with datasets.gdp6h() as ds:
            self.assertTrue(ds)

    def test_glad(self):
        with datasets.glad() as ds:
            self.assertTrue(ds)

    def test_glad_dims_coords(self):
        with datasets.glad() as ds:
            self.assertTrue(len(ds.sizes) == 2)
            self.assertTrue("obs" in ds.dims)
            self.assertTrue("traj" in ds.dims)
            self.assertTrue(len(ds.coords) == 2)
            self.assertTrue("time" in ds.coords)
            self.assertTrue("id" in ds.coords)

    def test_glad_subset_and_apply_ragged_work(self):
        with datasets.glad() as ds:
            ds_sub = subset(
                ds,
                {"id": ["CARTHE_001", "CARTHE_002"]},
                id_var_name="id",
                row_dim_name="traj",
            )
            self.assertTrue(ds_sub)
            mean_lon = apply_ragged(np.mean, [ds_sub.longitude], ds_sub.rowsize)
            self.assertTrue(mean_lon.size == 2)

    def test_spotters_opens(self):
        with datasets.spotters() as ds:
            self.assertTrue(ds)

    def test_subsurface_floats_opens(self):
        with datasets.subsurface_floats() as ds:
            self.assertTrue(ds)

    def test_andro_opens(self):
        with datasets.andro() as ds:
            self.assertTrue(ds is not None)
            self.assertTrue(len(ds.variables) > 0)
            self.assertTrue(len(ds["lon_d"]) > 0)

            self.assertTrue(
                len(ds.lat_d[np.logical_or(ds.lat_d > 90, ds.lat_d < -90)]) == 0
            )
            self.assertTrue(
                len(ds.lat_d[np.logical_or(ds.lon_d > 180, ds.lon_d < -180)]) == 0
            )

    def test_yomaha_opens(self):
        with datasets.yomaha() as ds:
            self.assertTrue(ds)

    def test_mosaic_opens(self):
        with datasets.mosaic() as ds:
            self.assertTrue(ds)



import sys
import unittest
from unittest.mock import patch

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np

from clouddrift.plotting import plot_ragged

if __name__ == "__main__":
    unittest.main()


class plotting_tests(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        """
        Create trajectories example
        """
        self.lon = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        self.lat = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        self.rowsize = [3, 3, 4]

    def test_lonlatrowsize(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        lon_test = np.append(self.lon, 3)
        rowsize_test = np.append(self.rowsize, 3)
        with self.assertRaises(ValueError):
            plot_ragged(ax, lon_test, self.lat, self.rowsize)
            plot_ragged(ax, self.lon, self.lat, rowsize_test)

    def test_axis(self):
        ax = 1
        with self.assertRaises(ValueError):
            plot_ragged(ax, self.lon, self.lat, self.rowsize)

    def test_plot_colored_trajectory(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        l = plot_ragged(
            ax, self.lon, self.lat, self.rowsize, colors=np.arange(len(self.rowsize))
        )
        self.assertIsInstance(l, plt.cm.ScalarMappable)

    def test_plot_colored_datapoints(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        l = plot_ragged(
            ax, self.lon, self.lat, self.rowsize, colors=np.arange(len(self.lat))
        )
        self.assertIsInstance(l, plt.cm.ScalarMappable)

    def test_plot_color_wrong_dimension(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        color_test = np.append(np.arange(len(self.lat)), 3)
        with self.assertRaises(ValueError):
            plot_ragged(ax, self.lon, self.lat, self.rowsize, colors=color_test)

    def test_plot_cartopy_transform(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())
        with self.assertRaises(ValueError):
            plot_ragged(
                ax,
                self.lon,
                self.lat,
                self.rowsize,
                colors=np.arange(len(self.rowsize)),
            )

    def test_plot_cartopy(self):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection=ccrs.Mollweide())
        l = plot_ragged(
            ax,
            self.lon,
            self.lat,
            self.rowsize,
            colors=np.arange(len(self.rowsize)),
            transform=ccrs.PlateCarree(),
        )
        self.assertIsInstance(l, plt.cm.ScalarMappable)

    def test_plot_segments(self):
        self.lon = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        self.lat = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        self.rowsize = [3, 3, 4]

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection=ccrs.Mollweide())
        l = plot_ragged(
            ax,
            self.lon,
            self.lat,
            self.rowsize,
            colors=np.arange(len(self.rowsize)),
            transform=ccrs.PlateCarree(),
        )
        self.assertIsInstance(l, plt.cm.ScalarMappable)

    def test_plot_segments_split(self):
        self.lon = [-170, -175, -180, 175, 170]
        self.lat = [0, 1, 2, 3, 4]
        self.rowsize = [5]

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection=ccrs.Mollweide())
        l = plot_ragged(
            ax,
            self.lon,
            self.lat,
            self.rowsize,
            colors=np.arange(len(self.rowsize)),
            transform=ccrs.PlateCarree(),
        )
        self.assertIsInstance(l, plt.cm.ScalarMappable)

    def test_plot_segments_split_domain(self):
        self.lon = [-1, -2, -3, 3, 2, 1]
        self.lat = [0, 1, 2, 3, 4, 5]
        self.rowsize = [6]

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection=ccrs.Mollweide())
        l = plot_ragged(
            ax,
            self.lon,
            self.lat,
            self.rowsize,
            colors=np.arange(len(self.rowsize)),
            transform=ccrs.PlateCarree(),
            tolerance=5,
        )
        self.assertIsInstance(l, plt.cm.ScalarMappable)

    def test_matplotlib_not_installed(self):
        try:
            del sys.modules["clouddrift.plotting"]
        except Exception as e:
            print(f"Could not delete module for testing purposes, error: {e}")

        with patch.dict(sys.modules, {"matplotlib": None}):
            with self.assertRaises(ImportError):
                from clouddrift.plotting import plot_ragged

                fig = plt.figure()
                ax = fig.add_subplot(1, 1, 1)
                plot_ragged(ax, self.lon, self.lat, self.rowsize)



import logging
from importlib.metadata import PackageNotFoundError, distribution

_logger = logging.getLogger(__name__)

try:
    version = distribution("clouddrift").version
except PackageNotFoundError:
    # Handle case when developing on a local copy of the library rather than using an
    # installed copy from a distribution.
    version = "9999"
    _logger.info(f"Cannot determine package version using {version}")
